import csv
import itertools
import re
import io
from typing import List, Tuple, Optional
from Bio.Seq import Seq


class Feature(object):
    """ 
    Represents a gene or CRISPR repeat array. This is used internally by :class:`operon_analyzer.genes.Operon`, 
    but appears in the auto-generated documentation for reference. 
    """

    def __init__(self,
                 name: str,
                 coordinates: Tuple[int, int],
                 orfid: str,
                 strand: Optional[int],
                 accession: str,
                 e_val: Optional[float],
                 description: str,
                 sequence: str,
                 bit_score: Optional[int] = None,
                 raw_score: Optional[int] = None,
                 aln_len: Optional[int] = None,
                 pident: Optional[float] = None,
                 nident: Optional[float] = None,
                 mismatch: Optional[float] = None,
                 positive: Optional[float] = None,
                 gapopen: Optional[int] = None,
                 gaps: Optional[int] = None,
                 ppos: Optional[float] = None,
                 qcovhsp: Optional[int] = None):
        # Note: for CRISPR repeats, the pipeline does not identify the strand, and
        # since BLAST was not used, there is no e-value, so these values are set to
        # None.
        self.name = name
        self.coordinates = coordinates
        assert self.start < self.end
        self.orfid = orfid
        self.strand = strand
        self.accession = accession
        self.e_val = e_val
        self.bit_score = bit_score
        self.description = description
        self.sequence = sequence
        self.raw_score = raw_score
        self.aln_len = aln_len
        self.pident = pident
        self.nident = nident
        self.mismatch = mismatch
        self.positive = positive
        self.gapopen = gapopen
        self.gaps = gaps
        self.ppos = ppos
        self.qcovhsp = qcovhsp
        self.ignored_reasons = []

    @property
    def start(self) -> int:
        return self.coordinates[0]

    @property
    def end(self) -> int:
        return self.coordinates[1]

    def __hash__(self):
        return hash(f"{self.name}{self.start}{self.end}{self.sequence}")

    def __eq__(self, other: 'Feature'):
        return self.name == other.name and \
               self.start == other.start and \
               self.end == other.end and \
               self.orfid == other.orfid and \
               self.strand == other.strand and \
               self.accession == other.accession and \
               self.e_val == other.e_val and \
               self.bit_score == other.bit_score and \
               self.description == other.description and \
               self.sequence == other.sequence and \
               self.raw_score == other.raw_score and \
               self.aln_len == other.aln_len and \
               self.pident == other.pident and \
               self.nident == other.nident and \
               self.mismatch == other.mismatch and \
               self.positive == other.positive and \
               self.gapopen == other.gapopen and \
               self.gaps == other.gaps and \
               self.ppos == other.ppos and \
               self.qcovhsp == other.qcovhsp

    def ignore(self, reason: str):
        self.ignored_reasons.append(reason)

    def __len__(self):
        """ Length of feature in nucleotides """
        return self.end - self.start

    def __repr__(self) -> str:
        return f"<Feature {self.name} {self.start}..{self.end}>"


class Operon(object):
    """
    Provides access to features that were found in the same genomic region,
    which presumably comprise an actual operon. Whether this is true in reality
    must be determined by the user, if that is meaningful to them.
    """

    def __init__(self,
                 contig: str,
                 contig_filename: str,
                 start: int,
                 end: int,
                 features: List[Feature]):
        assert contig, "Missing contig name"
        assert 0 <= start and 0 <= end, "Invalid contig position"
        assert features, "Contig did not contain any features"
        self.contig = contig
        self.contig_filename = contig_filename
        self.start = start
        self.end = end
        self._features = features
        self._sequence = None

    def __hash__(self):
        feature_data = "".join([f"{feature.name}{feature.start}{feature.end}{feature.sequence}" for feature in self._features])
        return hash(f"{self.contig}{self.start}{self.end}{feature_data}")

    def __eq__(self, other):
        return self.contig == other.contig and \
               self.start == other.start and \
               self.end == other.end and \
               len(set(tuple(itertools.chain(self._features, other._features)))) == len(self._features)

    def set_sequence(self, sequence: Seq):
        """ Stores the nucleotide sequence of the operon. """
        self._sequence = sequence

    @property
    def feature_region(self) -> Tuple[int, int]:
        lower_bound, upper_bound = self.end, self.start
        for feature in self:
            lower_bound = min(feature.start, lower_bound)
            upper_bound = max(feature.end, upper_bound)
        return lower_bound, upper_bound

    @property
    def feature_region_sequence(self) -> str:
        """ Returns the nucleotide sequence of the operon, excluding the regions outside of the outermost Features. """
        lower_bound, upper_bound = self.feature_region
        return self._sequence[lower_bound:upper_bound]

    @property
    def sequence(self) -> str:
        return self._sequence

    @property
    def all_genes(self):
        """ Iterates over all genes (i.e. not CRISPR arrays) in the operon regardless of whether it's been ignored. """
        yield from (feature for feature in self._features if feature.e_val is not None)

    @property
    def all_features(self):
        """ Iterates over all features in the operon regardless of whether it's been ignored. """
        yield from self._features

    def __iter__(self):
        """ Iterates over all features that haven't been judged to be unrelated to the operon. """
        yield from (feature for feature in self._features if not feature.ignored_reasons)

    def __len__(self):
        """ The number of relevant genes or CRISPR arrays in the operon. """
        return len(tuple(iter(self)))

    @property
    def feature_names(self):
        """ Iterates over the name of each feature in the operon """
        yield from (feature.name for feature in self)

    def get(self, feature_name: str, regex=False) -> List[Feature]:
        """ Returns a list of every Feature with a given name. """
        if regex:
            rx = re.compile(feature_name, re.IGNORECASE)
        else:
            rx = re.compile(f"^{feature_name}$", re.IGNORECASE)
        features = []
        for feature in self:
            if rx.search(feature.name):
                features.append(feature)
        return features

    def get_unique(self, feature_name: str, regex=False) -> Optional[Feature]:
        """ Returns a Feature or None if there is more than one Feature with the same name """
        features = self.get(feature_name, regex)
        if len(features) != 1:
            return None
        return features[0]

    def as_str(self) -> str:
        """
        Writes an Operon back out in the same CSV format that gene_finder produces. The text won't be
        completely identical in the case where floats have trailing decimals, or zero values in
        scientific format are recast as a simple float.
        """
        def optional_float(value, decimals: int) -> str:
            return f"{value:.{decimals}f}" if value else ""

        def optional_evalue(value) -> str:
            if value is None:
                return ""
            return f"{value:.2e}"

        def optional_float_or_int(value) -> str:
            if value is None:
                return ""
            if round(value) == value:
                return str(int(value))
            return f"{value:.1f}"

        def optional(value) -> str:
            return value if value not in ("", None) else ""

        output = io.StringIO(newline='')
        writer = csv.writer(output)
        for feature in self.all_features:
            feature_coords = f"{feature.start + 1}..{feature.end}" if feature.strand == 1 else f"{feature.end}..{feature.start + 1}"
            strand = feature.strand if feature.name != 'CRISPR array' else ''
            writer.writerow([f"{self.contig}",f"{self.start}..{self.end}",f"{feature.name}",f"{feature_coords}",f"{feature.orfid}",f"{optional(strand)}",f"{feature.accession}",f"{optional_evalue(feature.e_val)}",f"{feature.description}",f"{feature.sequence}",f"{optional_float_or_int(feature.bit_score)}",f"{optional(feature.raw_score)}",f"{optional(feature.aln_len)}",f"{optional_float(feature.pident, 3)}",f"{optional(feature.nident)}",f"{optional(feature.mismatch)}",f"{optional(feature.positive)}",f"{optional(feature.gapopen)}",f"{optional(feature.gaps)}",f"{optional_float(feature.ppos, 2)}",f"{optional(feature.qcovhsp)}",f"{self.contig_filename}"])
        return "%s" % output.getvalue().replace("\r", "")



import os, json, csv


def parse_blastn_output(tsv, step_id, parse_descriptions=True):
    """
    Parse output from a search step (in BLAST tabular format).

    Expects a tsv file with the following format fields: 
        qseqid sseqid stitle evalue \
        bitscore score length pident \
        nident mismatch positive gapopen \
        gaps ppos qcovhsp qseq qstart qend sstrand

    Returns a dictionary of best hits for each query that had a 
    hit, where "best" means the lowest e-value score.
    """
    try:
        hits = {}
        with open(tsv, newline='') as tsvfile:
            reader = csv.reader(tsvfile, delimiter="\t")

            for row in reader:
                if _keep_row(row, hits):
                    hit_dic = {}
                    hit_dic['type'] = 'nucleotide'

                    # information about the reference protein
                    hit_acc = row[1].strip()
                    hit_dic["Hit_accession"] = hit_acc if len(hit_acc) > 0 else "No accession/ID found for reference"
                    hit_def = row[2].split()
                    if parse_descriptions and len(hit_def) >= 2:
                        hit_dic["Hit_name"] = hit_def[1]
                        hit_dic["Hit_description"] = " ".join(hit_def)
                    else:
                        hit_def = " ".join(hit_def)
                        hit_dic["Hit_name"] = hit_def if len(hit_def) > 0 else "No gene name found for reference"
                        hit_dic["Hit_description"] = hit_def if len(hit_def) > 0 else "No gene description found for reference"
                    
                    # alignment statistics
                    hit_dic["Hit_e-val"] = row[3]
                    hit_dic["Bitscore"] = row[4]
                    hit_dic["Raw_score"] = row[5]
                    hit_dic["Alignment_length"] = row[6]
                    hit_dic["Alignment_percent-identical"] = row[7]
                    hit_dic["Alignment_num-identical"] = row[8]
                    hit_dic["Alignment_mismatches"] = row[9]
                    hit_dic["Alignment_num-positive"] = row[10]
                    hit_dic["Alignment_num-gapopenings"] = row[11]
                    hit_dic["Alignment_num-gaps"] = row[12]
                    hit_dic["Alignment_percent-pos"] = row[13]
                    hit_dic["Alignment_query-cov"] = row[14]
                    hit_dic["Query_seq"] = row[15]
                    hit_dic["Query_start-pos"] = row[16]
                    hit_dic["Query_end-pos"] = row[17]
                    hit_dic["Strand"] = 1 if row[18] == 'plus' else -1
                    hit_name = row[0]
                    hits[hit_name] = hit_dic

        hits = _reformat_hit_ids(hits, step_id)
        return hits
    
    except csv.Error:
        return {}


def parse_search_output(tsv, step_id, search_type, parse_descriptions=True):
    """
    Parse output from a search step (in BLAST tabular format).

    Expects a tsv file with the following format
    fields: qseqid sseqid stitle evalue \
        bitscore score length pident \
        nident mismatch positive (blast/diamond only) \
        gapopen gaps (blast/diamond only) \
        ppos (blast/diamond only) qcovs

    Returns a dictionary of best hits for each query that had a 
    hit, where "best" means the lowest e-value score.
    """
    try:
        hits = {}
        with open(tsv, newline='') as tsvfile:
            reader = csv.reader(tsvfile, delimiter="\t")

            for row in reader:
                if _keep_row(row, hits):
                    hit_dic = {}
                    local_query_id = row[0].split()[0]
                    hit_dic['type'] = 'protein'
                    hit_dic["Query_ORFID"] = local_query_id

                    # get query start/stop pos (nt) from ORF id
                    local_query_id = local_query_id.split("|")
                    hit_dic["Query_start-pos"] = local_query_id[1]
                    hit_dic["Query_end-pos"] = local_query_id[2]

                    # information about the reference protein
                    hit_acc = row[1].strip()
                    hit_dic["Hit_accession"] = hit_acc if len(hit_acc) > 0 else "No accession/ID found for reference"
                    hit_def = row[2].split()
                    if parse_descriptions and len(hit_def) >= 2:
                        hit_dic["Hit_name"] = hit_def[1]
                        hit_dic["Hit_description"] = " ".join(hit_def)
                    else:
                        hit_def = " ".join(hit_def)
                        hit_dic["Hit_name"] = hit_def if len(hit_def) > 0 else "No gene name found for reference"
                        hit_dic["Hit_description"] = hit_def if len(hit_def) > 0 else "No gene description found for reference"
                    
                    # alignment statistics
                    hit_dic["Hit_e-val"] = row[3]
                    hit_dic["Bitscore"] = row[4]
                    hit_dic["Raw_score"] = row[5]
                    hit_dic["Alignment_length"] = row[6]
                    hit_dic["Alignment_percent-identical"] = row[7]
                    hit_dic["Alignment_num-identical"] = row[8]
                    hit_dic["Alignment_mismatches"] = row[9]

                    if search_type in ("blast", "diamond"):
                        hit_dic["Alignment_num-positive"] = row[10]
                        hit_dic["Alignment_num-gapopenings"] = row[11]
                        hit_dic["Alignment_num-gaps"] = row[12]
                        hit_dic["Alignment_percent-pos"] = row[13]
                        hit_dic["Alignment_query-cov"] = row[14]
                        hit_dic["Query_seq"] = row[15]
                    
                    else: # mmseqs doesn't output some of these
                        hit_dic["Alignment_num-positive"] = None
                        hit_dic["Alignment_num-gapopenings"] = row[10]
                        hit_dic["Alignment_num-gaps"] = None
                        hit_dic["Alignment_percent-pos"] = None
                        hit_dic["Alignment_query-cov"] = row[11]
                        hit_dic["Query_seq"] = row[12]

                    hit_name = row[0]
                    hits[hit_name] = hit_dic

        hits = _reformat_hit_ids(hits, step_id)
        return hits
    
    except csv.Error:
        return {}


def _is_int(string):
    """
    Check if a string can be cast to an int.
    """
    try: 
        int(string)
        return True
    except ValueError:
        return False


def _get_column_lables(line):
    """
    Extract the following column lables from 
    pilercr "SUMMARY BY POSITION" table: 
    "Position", "Length", "Copies", "Repeat", "Spacer", "Strand",
    "Consensus".
    """
    lables = line.split()
    lables_to_remove = ["Array", "Sequence", "#", "+"]
    lables = [lable for lable in lables if lable not in lables_to_remove]
    lables.insert(5, "Strand")
    
    return lables
 

def _get_array_info(line, array_num):
    """Extract the following information for an array
    in pilercr "SUMMARY BY SIMILARITY" table: 
    "Position", "Length", "Copies", "Repeat", "Spacer", "Strand",
    "Consensus".
    """
    line = line[25:]
    array_info = line.split()
    array_id = "Array_{}".format(array_num)
    
    return array_id, array_info


def parse_pilercr_summary(pilercr_out):
    """
    Parse pilercr output in the "summary by position" section only.

    Returns a dictionary of dictionaries, containing
    key-value pairs of infomation associated with
    each array identified by piler.

    Example:

        >>> arrays = parse_pilercr_summary("vcrass_pilercr_out")
        >>> print(json.dumps(arrays, indent=4))
        {
            "array_1": {
                "Position": "30001",
                "Length": "148",
                "Copies": "3",
                "Repeat": "28",
                "Spacer": "32",
                "Consensus": "GTGAACTGCCGAATAGGTAGCTGATAAT"
            }
        }

    """    
    headers_reached = False
    array_info_start = False
    lables = []
    arrays = {}
    array_num = 0

    with open(pilercr_out) as f:
        for line in f:

            if headers_reached and not line.isspace():
                if line.strip().split()[0] == "Array":
                    lables = _get_column_lables(line)
                    array_info_start = True
                    headers_reached = False

            elif array_info_start and not line.isspace():
                if _is_int(line.strip().split()[0]):
                    array_id, array_info = _get_array_info(line, array_num)
                    arrays[array_id] = {}
                    array_num += 1
                    for lable, value in zip(lables, array_info):
                        arrays[array_id][lable] = value

            if line.strip() == "SUMMARY BY SIMILARITY":
                headers_reached = True

            if line.strip() == "SUMMARY BY POSITION":
                array_info_start = False

    return arrays


def _keep_row(row, hits):
    """
    Determine if given row represents the best hit for this
    query so far by comparing e-values. 
    """
    if row[0] in hits:
        if float(row[3]) < float(hits[row[0]]["Hit_e-val"]):
            return True
        else:
            return False
    else:
        return True


def _reformat_hit_ids(hits, step_id):
    """
    For consistency, copy over hit info from a dict where keys are the
    utility-specific query IDs to a dict where keys are in the 
    format created by the other gene finder parsers.
    """
    re_hits = {}
    hit_num = 0
    for hit in hits.values():
        hit_id = "{}{}_hit-{}".format(step_id[:1].upper(), step_id[1:], str(hit_num))
        re_hits[hit_id] = hit
        hit_num += 1
    
    return re_hits



import os, shutil

def concatenate(in_dir, file_names):
    """
    Concatenate two or more fasta files.
    """
    out = os.path.join(in_dir, "merged_input.fasta")
    if os.path.exists(out):
        os.remove(out)
    with open(out,"wb") as outfile:
        for name in file_names:
            with open(name,"rb") as infile:
                shutil.copyfileobj(infile, outfile) 
    return out

    
def get_neighborhood_ranges(hits, contig_len, span=20000):
    """
    For each BLAST hit in `hits`, get the coordinates of the surrounding genomic neighborhood, which
    is the area directly `span` bp up- and downstream of each hit. When two or more hits have 
    overlapping neighborhoods, merge them into a single region.
    """
    # sort the hit dictionary keys by hit start position
    keys_sorted = sorted(hits, key=lambda k: min(int(hits[k]["Query_start-pos"]), int(hits[k]["Query_end-pos"])))
    
    tmp_ranges = []
    # use sorted keys to construct a temporary list of ranges
    for key in keys_sorted:
        start = int(hits[key]["Query_start-pos"])
        stop = int(hits[key]["Query_end-pos"])
        lower = max(min((start - span), (stop - span)), 0)
        upper = min(max((start + span), (stop + span)), contig_len)
        tmp_ranges.append((lower , upper))
    
    final = []
    # merge overlapping ranges to create final list
    for this_coord in tmp_ranges:
        if len(final) == 0:
            final.append(this_coord)
            continue
        
        last_coord = final[-1]
        if this_coord[0] > last_coord[1]:
            final.append(this_coord)
        elif this_coord[0] <= last_coord[1] and this_coord[1] > last_coord[1]:
            final[-1] = (last_coord[0], this_coord[1])
        else:
            continue

    return final



import csv
from gene_finder.output_writers import FIELDNAMES
from typing import Tuple, Iterator, IO
from operon_analyzer.genes import Feature, Operon
import sys


csv.field_size_limit(sys.maxsize)

PipelineRecord = Tuple[str, str, str, str, str, str, str, str, str, str]
Coordinates = Tuple[int, int]


def assemble_operons(lines: Iterator[PipelineRecord]) -> Iterator[Operon]:
    """
    Takes the output from :class:`gene_finder.pipeline.Pipeline` and loads all features,
    then assembles them into :class:`operon_analyzer.genes.Operon` objects.

    To keep things memory efficient while not allowing redundant operons from being loaded,
    we keep track of the hash of each operon, which is just an integer. Even for very
    large metagenomic databases this should use less than a gigabyte of memory.

    This is a helper function used by the loaders in :mod:`operon_analyzer.load`, but
    appears in the auto-generated documentation for reference.
    """
    hashes = set()
    first = next(lines)
    contig, contig_filename, coordinates, feature = _parse_feature(first)
    current_contig = (contig, contig_filename, coordinates)
    features = [feature]

    for line in lines:
        if not line:
            continue
        contig, contig_filename, coordinates, feature = _parse_feature(line)
        if (contig, contig_filename, coordinates) != current_contig:
            (current_contig, current_contig_filename, (current_start, current_end)) = current_contig
            operon = Operon(current_contig, current_contig_filename, current_start, current_end, features)
            h = hash(operon)
            if h not in hashes:
                hashes.add(h)
                yield operon
            features = [feature]
            current_contig = (contig, contig_filename, coordinates)
        else:
            features.append(feature)
    (current_contig, current_contig_filename, (current_start, current_end)) = current_contig
    operon = Operon(current_contig, current_contig_filename, current_start, current_end, features)
    h = hash(operon)
    if h not in hashes:
        hashes.add(h)
        yield operon


def parse_coordinates(coordinates: str) -> Coordinates:
    """ Parses base pair ranges denoting position in a contig.
    Since BLAST and piler-cr both use 1-based indices, we subtract 1
    from the start coordinate since Opfi internally uses 0-based indices. """
    start, end = coordinates.split("..")
    return int(start), int(end)


def read_pipeline_output(handle: IO[str]) -> Iterator[PipelineRecord]:
    """ Reads the CSV file produced by the CRISPR-transposon pipeline """
    reader = csv.reader(handle)
    for record in reader:
        if record == FIELDNAMES:
            continue
        yield record


def _parse_feature(line: PipelineRecord) -> Tuple[str, Coordinates, Feature]:
    """ Creates a Feature from a line of output from a CSVReader """
    contig = line[0]
    coordinates = parse_coordinates(line[1])
    feature = line[2]

    # Piler-cr and BLAST both use 1-based indices, but Opfi uses 0-based indices.
    # To make both coordinate systems consistent, we subtract 1 from the start
    # since feature coordinates come directly from those tools.
    # If features are on the reverse strand, the second coordinate will be larger
    # than the first, but operon_analyzer assumes the start is always less than the
    # end
    first_coord, second_coord = parse_coordinates(line[3])
    feature_start = min(first_coord, second_coord) - 1
    feature_end = max(first_coord, second_coord)

    query_orfid = line[4]
    strand = int(line[5]) if line[5] else (1 if feature_start < feature_end else -1)
    hit_accession = line[6]
    hit_eval = float(line[7]) if line[7] else None
    description = line[8]
    sequence = line[9]
    if len(line) > 10:
        bit_score = float(line[10]) if line[10] != '' else None
        raw_score = int(line[11]) if line[11] != '' else None
        aln_len = int(line[12]) if line[12] != '' else None
        pident = float(line[13]) if line[13] != '' else None
        nident = int(line[14]) if line[14] != '' else None
        mismatch = int(line[15]) if line[15] != '' else None
        positive = int(line[16]) if line[16] != '' else None
        gapopen = int(line[17]) if line[17] != '' else None
        gaps = int(line[18]) if line[18] != '' else None
        ppos = float(line[19]) if line[19] != '' else None
        qcovhsp = int(line[20]) if line[20] != '' else None
        contig_filename = line[21] if line[21] else ''
    else:
        bit_score = None
        raw_score = None
        aln_len = None
        pident = None
        nident = None
        mismatch = None
        positive = None
        gapopen = None
        gaps = None
        ppos = None
        qcovhsp = None
        contig_filename = None

    return contig, contig_filename, coordinates, Feature(
        feature,
        (feature_start, feature_end),
        query_orfid,
        strand,
        hit_accession,
        hit_eval,
        description,
        sequence,
        bit_score,
        raw_score,
        aln_len,
        pident,
        nident,
        mismatch,
        positive,
        gapopen,
        gaps,
        ppos,
        qcovhsp)



from collections import defaultdict
from typing import Tuple, Dict, DefaultDict, Iterator, IO, List
from operon_analyzer.analyze import load_analyzed_operons


def load_counts(lines: IO[str]) -> Tuple[Dict[str, int],
                                         Dict[str, int],
                                         Dict[int, int]]:
    """ Takes a stream of the output from :func:`operon_analyzer.analyze.analyze` and computes some useful statistics. Namely,
    the number of times each rule was the only one broken for a contig, the number of times each rule was broken regardless of context,
    and the occurrences of the count of broken rules per contig. """
    operons = load_analyzed_operons(lines)
    results = _extract_results(operons)
    unique_rule_violated, failed_rule_occurrences, rule_failure_counts = _count_results(results)
    return unique_rule_violated, failed_rule_occurrences, rule_failure_counts


def _extract_results(lines: Iterator[Tuple[str, int, int, str]]) -> Iterator[str]:
    """ Throws out data we don't care about from the analysis output and returns
    only the actual results """
    for _, _, _, _, result in lines:
        yield result


def _count_results(results: Iterator[List[str]]) -> Tuple[Dict[str, int],
                                                          Dict[str, int],
                                                          Dict[int, int]]:
    """
    Returns a tuple of:
      - the number of times each rule was the only one broken for a contig
      - the number of times each rule was broken regardless of context
      - the occurrences of the count of broken rules per contig
    """

    unique_rule_violated = defaultdict(int)  # type: DefaultDict[str, int]
    failed_rule_occurrences = defaultdict(int)  # type: DefaultDict[str, int]
    rule_failure_counts = defaultdict(int)  # type: DefaultDict[int, int]
    all_rules = set()

    for result in results:
        # result will either be the singular word "pass", or the word "fail" followed
        # by space-separated rules that the operon broke
        if result[0] == 'pass':
            # there were no failed rules
            rule_failure_counts[0] += 1
        else:
            # there was at least one failed rule
            assert result[0] == 'fail'
            for rule in result[1:]:
                failed_rule_occurrences[rule] += 1
                all_rules.add(rule)
            if len(result) == 2:
                # there was exactly one failed rule
                unique_rule_violated[rule] += 1
            # keep track of how many failed rules there were
            rule_failure_counts[len(result[1:])] += 1

    # Ensure rules with zero counts are represented in the data
    for rule in all_rules:
        failed_rule_occurrences[rule] += 0
        unique_rule_violated[rule] += 0
    # convert all results into regular dictionaries since leaving them as defaultdicts
    # could only increase the possibility of a downstream bug
    return dict(unique_rule_violated), dict(failed_rule_occurrences), dict(rule_failure_counts)



BLAST_OPTIONS_COMMON = ['dbsize',
'word_size', 'gapopen', 'gapextend', 'qcov_hsp_perc', 
'xdrop_ungap', 'xdrop_gap', 'xdrop_gap_final', 'searchsp', 
'sum_stats', 'seg', 'soft_masking', 'matrix', 
'threshold', 'culling_limit', 'window_size', 'num_threads', 
'comp_based_stats', 'gilist', 'seqidlist', 'negative_gilist',
'db_soft_mask', 'db_hard_mask', 'entrez_query', 'max_hsps',
'best_hit_overhang',  'best_hit_score_edge', 'max_target_seqs',
'import_search_strategy', 'export_search_strategy', 'num_alignments']

BLAST_FLAGS = ['lcase_masking', 'ungapped', 'use_sw_tback', 'remote']

PSIBLAST_FLAGS = ['lcase_masking', 'use_sw_tback', 
'save_pssm_after_last_round', 'save_each_pssm', 'remote']

BLASTP_OPTIONS = ['task']

PSIBLAST_OPTIONS = ['gap_trigger', 'num_iterations', 'out_pssm', 
'out_ascii_pssm', 'pseudocount', 'inclusion_ethresh']

BLASTN_OPTIONS = ['filtering_algorithm', 'sum_stats', 'window_masker_db', 'window_size', 'template_type', 'version', 'parse_deflines', 'min_raw_gapped_score', 'string', 'format', 'max_hsps', 'taxids', 'negative_taxids', 'num_alignments', 'strand', 'off_diagonal_range', 'subject_besthit', 'num_sequences', 'no_greedy', 'negative_taxidlist', 'culling_limit', 'xdrop_ungap', 'open_penalty', 'DUST_options', 'sorthits', 'xdrop_gap_final', 'negative_gilist', 'subject', 'use_index', 'bool_value', 'filename', 'seqidlist', 'task_name', 'sort_hits', 'database_name', 'lcase_masking', 'query_loc', 'subject_loc', 'sort_hsps', 'line_length', 'boolean', 'db_hard_mask', 'negative_seqidlist', 'template_length', 'filtering_db', 'filtering_database', 'penalty', 'searchsp', 'ungapped', 'type', 'gapextend', 'db_soft_mask', 'dbsize', 'qcov_hsp_perc', 'sorthsps', 'window_masker_taxid', 'index_name', 'export_search_strategy', 'float_value', 'soft_masking', 'gilist', 'entrez_query', 'show_gis', 'best_hit_score_edge', 'gapopen', 'subject_input_file', 'range', 'html', 'word_size', 'best_hit_overhang', 'perc_identity', 'input_file', 'num_descriptions', 'xdrop_gap', 'dust', 'taxidlist', 'max_target_seqs', 'num_threads', 'task', 'remote', 'int_value', 'extend_penalty', 'reward', 'import_search_strategy', 'num_letters']


def build_blastp_command(query, db, evalue, kwargs, output_fields, out, blastp_path):
    """
    Generate a python subprocess command (list) for
    executing blastp.

    kwargs should be valid blastp options. Note that
    at this time, only certain options are allowed.
    """

    cmd = [blastp_path, "-query", query, "-db", db, "-evalue", str(evalue), "-out", out]
    for key, value in kwargs.items():
        if key in (BLAST_OPTIONS_COMMON + BLASTP_OPTIONS):
            cmd.append("-{}".format(key))
            cmd.append(str(value))
        elif key in (BLAST_FLAGS) and value:
            cmd.append("-{}".format(key))

    cmd.append("-outfmt")
    cmd.append("6 {}".format(output_fields))
    return cmd

def build_psiblast_command(query, db, evalue, kwargs, output_fields, out, psiblast_path):
    """
    Generate a python subprocess command (list) for
    executing psiblast.

    kwargs should be valid blastp options. Note that
    at this time, only certain options are allowed.
    """

    cmd = [psiblast_path, "-query", query, "-db", db, "-evalue", str(evalue), "-out", out]
    for key, value in kwargs.items():
        if key in (BLAST_OPTIONS_COMMON + PSIBLAST_OPTIONS):
            cmd.append("-{}".format(key))
            cmd.append(str(value))
        elif key in (PSIBLAST_FLAGS) and value:
            cmd.append("-{}".format(key))

    cmd.append("-outfmt")
    cmd.append("6 {}".format(output_fields))
    return cmd

def build_blastn_command(query, db, evalue, kwargs, output_fields, out, blastn_path):
    """
    Generate a python subprocess command (list) for
    executing blastn.

    kwargs should be valid blastn options. 
    """

    cmd = [blastn_path, "-query", query, "-db", db, "-evalue", str(evalue), "-out", out]
    for key, value in kwargs.items():
        if key in BLASTN_OPTIONS:
            cmd.append("-{}".format(key))
            cmd.append(str(value))

    cmd.append("-outfmt")
    cmd.append("6 {}".format(output_fields))
    return cmd



import csv

# Header values in old pipeline output
# CSVWriter no longer writes headers, but this list
# is still needed for operon_analyzer to handle
# old output formats
FIELDNAMES = ["Contig", "Locus coordinates", "Feature", 
              "Feature coordinates", "Query ORFID",
              "Strand", "Hit accession", "Hit e-val", 
              "Description", "Query/Consensus sequence"]


class CSVWriter:
    """
    Write pipeline output to a CSV formatted file.
    """
    def __init__(self, results, outfile):
        """
        Args:
            results (dict): All candidates identified by the pipeline.
            outfile (string): The file to write output to.
        """
        self.results = results
        self.id = None
        self.project_id = None
        self.outfile = outfile
    

    def _ret_fieldnames(self):
        # is this still needed?
        return FIELDNAMES
    

    def _get_nucleotide_row(self, neighborhood, hit):
        """
        Format a row representing a hit identified with nucleotide BLAST.
        """
        row = [""] * 22
        row[0] = self.id
        row[1] = "{}..{}".format(neighborhood["Loc_start-pos"],
                                neighborhood["Loc_end-pos"])
        row[2] = hit["Hit_name"]
        row[3] = "{}..{}".format(hit["Query_start-pos"],
                                hit["Query_end-pos"])
        row[4] = ''
        row[5] = hit["Strand"]
        row[6] = hit["Hit_accession"]
        row[7] = hit["Hit_e-val"]
        row[8] = hit["Hit_description"]
        row[9] = hit["Query_seq"]
        row[10] = hit["Bitscore"]
        row[11] = hit["Raw_score"]
        row[12] = hit["Alignment_length"]
        row[13] = hit["Alignment_percent-identical"]
        row[14] = hit["Alignment_num-identical"]
        row[15] = hit["Alignment_mismatches"]
        row[16] = hit["Alignment_num-positive"] if hit["Alignment_num-positive"] is not None else ""
        row[17] = hit["Alignment_num-gapopenings"] 
        row[18] = hit["Alignment_num-gaps"] if hit["Alignment_num-gaps"] is not None else ""
        row[19] = hit["Alignment_percent-pos"] if hit["Alignment_percent-pos"] is not None else ""
        row[20] = hit["Alignment_query-cov"]
        row[21] = self.project_id

        return row
    

    def _get_row(self, neighborhood, hit):
        """
        Format a row representing a hit identified with protein BLAST (or psiBLAST, mmseqs, or 
        Diamond).
        """
        row = [""] * 22
        row[0] = self.id
        row[1] = "{}..{}".format(neighborhood["Loc_start-pos"],
                                neighborhood["Loc_end-pos"])
        row[2] = hit["Hit_name"]
        row[3] = "{}..{}".format(hit["Query_start-pos"],
                                hit["Query_end-pos"])
        row[4] = hit["Query_ORFID"]
        row[5] = hit["Query_ORFID"].split("|")[-1]
        row[6] = hit["Hit_accession"]
        row[7] = hit["Hit_e-val"]
        row[8] = hit["Hit_description"]
        row[9] = hit["Query_seq"]
        row[10] = hit["Bitscore"]
        row[11] = hit["Raw_score"]
        row[12] = hit["Alignment_length"]
        row[13] = hit["Alignment_percent-identical"]
        row[14] = hit["Alignment_num-identical"]
        row[15] = hit["Alignment_mismatches"]
        row[16] = hit["Alignment_num-positive"] if hit["Alignment_num-positive"] is not None else ""
        row[17] = hit["Alignment_num-gapopenings"] 
        row[18] = hit["Alignment_num-gaps"] if hit["Alignment_num-gaps"] is not None else ""
        row[19] = hit["Alignment_percent-pos"] if hit["Alignment_percent-pos"] is not None else ""
        row[20] = hit["Alignment_query-cov"]
        row[21] = self.project_id

        return row
    

    def _format_array_des(self, hit):
        """
        Create a string representation of a CRISPR array hit.
        """
        copy = hit["Copies"]
        repeat = hit["Repeat"]
        spacer = hit["Spacer"]

        array_des = "Copies: {}, Repeat: {}, Spacer: {}".format(copy, repeat, spacer)
        
        return array_des


    def _get_crispr_array_row(self, neighborhood, array):
        """
        Format a row representing a CRISPR array identified using PILER-CR.
        """
        row = [""] * 22
        row[0] = self.id
        row[1] = "{}..{}".format(neighborhood["Loc_start-pos"],
                                                    neighborhood["Loc_end-pos"])
        row[2] = "CRISPR array"
        row[3] = "{}..{}".format(array["Position"],
                                str(int(array["Position"]) + int(array["Length"])))
        row[8] = self._format_array_des(array)
        row[9] = array["Consensus"]
        row[21] = self.project_id

        return row
    

    def _get_rows(self, neighborhood):
        """
        Create a CSV entry for a candidate identified by the pipeline, where each gene/feature 
        is represented by a single row.
        """
        rows = []
        for hit in neighborhood["Hits"]:
            hit_type = neighborhood["Hits"][hit].get('type')
            if hit_type == 'protein':
                rows.append(self._get_row(neighborhood, neighborhood["Hits"][hit]))
            elif hit_type == 'nucleotide':
                rows.append(self._get_nucleotide_row(neighborhood, neighborhood["Hits"][hit]))
            else:
                rows.append(self._get_crispr_array_row(neighborhood, 
                                                        neighborhood["Hits"][hit]))
        
        return rows


    def to_csv(self, project_id, mode):
        """
        Write candidate CSV data to disk.

        Args:
            project_id (string): The filename/path for the input data used by this run 
                of the pipeline. 
            mode (string): The mode for opening the output file for writing. Usually this
                will be "w" (for create/write mode), but if the pipeline is being run 
                incrementally then results may need to be appended to the output file 
                periodically.
        """
        with open(self.outfile, mode, newline='') as csvfile:
            writer = csv.writer(csvfile)
            
            self.project_id = project_id
            for contig in self.results:
                self.id = contig
                for neighborhood in self.results[contig]:
                    rows = self._get_rows(self.results[contig][neighborhood])
                    for row in rows:
                        writer.writerow(row)



import multiprocessing
import os
import subprocess
import tempfile

from Bio.Blast.Applications import (NcbiblastpCommandline,
                                    NcbipsiblastCommandline)

from gene_finder.option_handling import (build_blastn_command,
                                         build_blastp_command,
                                         build_psiblast_command)
from gene_finder.parsers import (parse_blastn_output, parse_pilercr_summary,
                                 parse_search_output)
from gene_finder.utils import get_neighborhood_ranges


class Blastp():
    """
    Wrapper for the NCBI BLASTP command line utility.
    """
    BLASTOUT_FIELDS = "qseqid sseqid stitle evalue bitscore score length pident nident mismatch positive gapopen gaps ppos qcovhsp qseq"

    def __init__(self, db, e_val, step_id, parse_descriptions, blastp_path, kwargs):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.db = db
        self.e_val = e_val
        self.step_id = step_id
        self.parse_descriptions = parse_descriptions
        self.blastp_path = blastp_path
        self.kwargs = kwargs
    
    def construct_cmd(self, query, out):
        """
        Format the BLASTP command into a string that `subprocess.run()` can use.
        """
        return build_blastp_command(query, self.db, self.e_val, self.kwargs, self.BLASTOUT_FIELDS, out, self.blastp_path)

    def run(self, orfs):
        """
        Execute the BLASTP search.
        """
        blast_out = os.path.join(self.tmp_dir.name, "blast_out.tsv")
        cmd = self.construct_cmd(orfs, blast_out)
        subprocess.run(cmd, check=True)
        hits = parse_search_output(blast_out, self.step_id, "blast", self.parse_descriptions)
        return hits
        

class Blastpsi():
    """
    Wrapper for NCBI psiBLAST command line utility.
    """
    BLASTOUT_FIELDS = "qseqid sseqid stitle evalue bitscore score length pident nident mismatch positive gapopen gaps ppos qcovhsp qseq"

    def __init__(self, db, e_val, step_id, parse_descriptions, psiblast_path, kwargs):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.db = db
        self.e_val = e_val
        self.step_id = step_id
        self.parse_descriptions = parse_descriptions
        self.psiblast_path = psiblast_path
        self.kwargs = kwargs

    def construct_cmd(self, query, out):
        """
        Format the psiBLAST command into a string that `subprocess.run()` can use.
        """
        return build_psiblast_command(query, self.db, self.e_val, self.kwargs, self.BLASTOUT_FIELDS, out, self.psiblast_path)

    def run(self, orfs):
        """
        Execute the psiBLAST search.
        """
        blast_out = os.path.join(self.tmp_dir.name, "blast_out.tsv")
        cmd = self.construct_cmd(orfs, blast_out)
        subprocess.run(cmd, check=True)
        hits = parse_search_output(blast_out, self.step_id, "blast", self.parse_descriptions)
        return hits


class MMseqs():
    """
    Wrapper for mmseqs command line search utility.
    """
    MMSEQSOUT_FIELDS = "qheader target theader evalue bits raw alnlen pident nident mismatch gapopen qcov qseq"

    def __init__(self, db, e_val, step_id, sensitivity, parse_descriptions):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.target_db = db
        self.step_id = step_id
        self.e_val = e_val
        self.sensitivity = sensitivity
        self.parse_descriptions = parse_descriptions
    
    def _make_query_db(self, orfs):
        """
        Convert query orfs to mmseqs custom database format.
        """
        query_db = "{}/querydb".format(self.tmp_dir.name)
        createdb_cmd = ["mmseqs", "createdb", orfs, query_db]
        subprocess.run(createdb_cmd, check=True)
        return query_db
    
    def _extract_best_hits(self, query_db, result_db):
        """
        Extract best hit info from raw mmseqs output.
        For simplicity, "best" means the hit with the 
        lowest e-value score.
        """
        results_tsv = "{}/results.tsv".format(self.tmp_dir.name)
        tsv_fields = ",".join(self.MMSEQSOUT_FIELDS.split())

        # convert raw output from custom mmseqs format to tsv format
        convertalis_cmd = ["mmseqs", "convertalis", query_db, self.target_db, 
                            result_db, results_tsv, "--format-output", tsv_fields]
        subprocess.run(convertalis_cmd, check=True)
        hits = parse_search_output(results_tsv, self.step_id, "mmseqs", self.parse_descriptions)
        return hits
    
    def run(self, orfs):
        """
        Execute the mmseqs search and parse output.
        """
        # setup necessary files and directories for mmseqs run
        query_db = self._make_query_db(orfs)
        result_db = "{}/resultdb".format(self.tmp_dir.name)
        mmseqs_work = os.path.join(self.tmp_dir.name, "tmp")
        os.mkdir(mmseqs_work)

        cmd = ["mmseqs", "search", query_db, self.target_db, result_db, mmseqs_work, "-e", self.e_val, "-v", "0"]
        if self.sensitivity is not None:
            cmd.append("-s")
            cmd.append(str(self.sensitivity))
        subprocess.run(cmd, check=True)
        # parse output
        hits = self._extract_best_hits(query_db, result_db)
        return hits
    

class Diamond():
    """
    Wrapper for diamond command line search utility.
    """
    DIAMONDOUT_FIELDS = "qseqid sseqid stitle evalue bitscore score length pident nident mismatch positive gapopen gaps ppos qcovhsp qseq"

    def __init__(self, db, e_val, step_id, sensitivity, parse_descriptions):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.db = db
        self.step_id = step_id
        self.e_val = e_val
        self.sensitivity = sensitivity
        self.parse_descriptions = parse_descriptions

    def run(self, orfs):
        """
        Execute the diamond blastp search and parse output.
        """
        result = os.path.join(self.tmp_dir.name, "result.tsv")
        cmd = ['diamond', 'blastp', '-q', orfs, '-d', self.db, '-o', result,
                '-e', self.e_val, '--threads', '1', '--quiet', '-f', '6']
        cmd = cmd + self.DIAMONDOUT_FIELDS.split()
        if self.sensitivity is not None:
            cmd.append(str(self.sensitivity))
        subprocess.run(cmd, check=True)
        # parse output
        hits = parse_search_output(result, self.step_id, "diamond", self.parse_descriptions)
        return hits       


class Pilercr():
    """
    Wrapper for PILER-CR CRISPR array finding software.
    """
    def __init__(self, step_id):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.step_id = step_id
    
    def run(self, genome):
        """
        Execute the PILER-CR array search. 
        """
        pilercr_out = os.path.join(self.tmp_dir.name, "pilercr_results")
        cmd = ["pilercr", "-in", genome, "-out", pilercr_out, "-minarray", "2", "-quiet"]
        subprocess.run(cmd, check=True)
        hits = parse_pilercr_summary(pilercr_out)
        return hits
        

class Blastn():
    """
    Wrapper for NCBI blastn command line utility.
    """
    BLASTOUT_FIELDS = "qseqid sseqid stitle evalue bitscore score length pident nident mismatch positive gapopen gaps ppos qcovhsp qseq qstart qend sstrand"

    def __init__(self, db, step_id, e_val, parse_descriptions, blastn_path, kwargs):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.db = db
        self.step_id = step_id
        self.e_val = e_val
        self.parse_descriptions = parse_descriptions
        self.blastn_path = blastn_path
        self.kwargs = kwargs
    
    def construct_cmd(self, query, out):
        return build_blastn_command(query, self.db, self.e_val, self.kwargs, self.BLASTOUT_FIELDS, out, self.blastn_path)

    def run(self, genome):
        blast_out = os.path.join(self.tmp_dir.name, "blast_out.tsv")
        cmd = self.construct_cmd(genome, blast_out)
        subprocess.run(cmd, check=True)
        hits = parse_blastn_output(blast_out, self.step_id, self.parse_descriptions)
        return hits      


class SearchStep:
    """
    Represents a generic step in the pipeline.
    """
    def __init__ (self, search_tool):
        self.search_tool = search_tool

    def execute(self, orfs):
        self.orfs = orfs
        self.hits = self.search_tool.run(orfs)


class SeedStep(SearchStep):
    """
    Represents a bait/seed step in the pipeline.
    """
    def __init__ (self, search_tool):
        SearchStep.__init__(self, search_tool)
    
    def execute(self, orfs, span, contig_len):
        super().execute(orfs)
        if len(self.hits) != 0:
            self.neighborhood_ranges = get_neighborhood_ranges(self.hits, contig_len, span)
        else:
            self.neighborhood_ranges = []


class SeedWithCoordinatesStep():
    """
    Represents a bait/seed step in the pipeline, where coordinates are used to 
    define candidates.
    """
    def __init__(self, start, end, contig_id):
        # these shouldn't change after initialization
        self.start = start
        self.end = end
        self.contig_id = contig_id

        # Will be updated for each contig if no 
        # coordinates were given when the step was added
        self.coordinates = (start, end)
    
    def update_start_coord(self, new_start):
        self.coordinates = (new_start, self.coordinates[1])
    
    def update_end_coord(self, new_end):
        self.coordinates = (self.coordinates[0], new_end)
    
    @property
    def neighborhood_ranges(self):
        return [self.coordinates]


class FilterStep(SearchStep):
    """
    Represents a filtering step in the pipeline. 
    """
    def __init__ (self, search_tool, min_prot_count):
        SearchStep.__init__(self, search_tool)
        self.min_prot_count = min_prot_count


class CrisprStep(SearchStep):
    """
    Represents a CRISPR array search step in the pipeline.
    """
    def __init__(self, search_tool):
        SearchStep.__init__(self, search_tool)


class BlastnStep(SearchStep):
    """
    Represents a nucleotide BLAST search step in the pipeline.
    """
    def __init__(self, search_tool):
        SearchStep.__init__(self, search_tool)



import gzip
from typing import Optional, Iterator, IO
from Bio import SeqIO
from Bio.Seq import Seq
from operon_analyzer import genes, parse


def load_sequence(operon: genes.Operon) -> Optional[Seq]:
    """ Loads the DNA sequence for a given operon's contig. """
    with gzip.open(operon.contig_filename, 'rt') as f:
        records = SeqIO.parse(f, 'fasta')
        for record in records:
            if record.id == operon.contig:
                return Seq(str(record.seq))


def load_gzipped_operons(filename: str) -> Iterator[genes.Operon]:
    """
    Create a :class:`operon_analyzer.genes.Operon` representation for each candidate
    present in gzipped output from :class:`gene_finder.pipeline.Pipeline`.

    Args:
        filename (str): Path to the compressed pipeline CSV file. 
    """
    with gzip.open(filename, 'rt') as f:
        for operon in load_operons(f):
            yield operon


def load_operons(handle: IO[str]) -> Iterator[genes.Operon]:
    """
    Create a :class:`operon_analyzer.genes.Operon` representation for each candidate
    present in output from :class:`gene_finder.pipeline.Pipeline`.

    Args:
        handle (IO): Handle for the pipeline CSV file. 
    """
    yield from parse.assemble_operons(parse.read_pipeline_output(handle))



import re
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Tuple

from operon_analyzer import analyze, genes

ReannotationCounts = Dict[Tuple[str, int], Dict[str, int]]
ReannotationFractions = Dict[Tuple[str, int], Dict[str, float]]


def summarize(operons: List[genes.Operon], reannotated_operons: List[genes.Operon]):
    """ For each protein in each operon, prints out how frequently it was reannotated to a particular protein.
    For example, a putative Cas9 may be identified as a transposase when BLASTed with a more expansive database. """
    clusters, reblasted_operons = _prepare_operons_for_counting(operons, reannotated_operons)
    for label, cloperons in clusters.items():
        counts = _count_cluster_reannotations(cloperons, reblasted_operons)
        if not counts:
            continue
        fractions = _convert_reannotation_counts_to_fractions(counts)
        for (feature, count), reannotations in fractions.items():
            print(_format_reannotation_summary("-".join(label), len(cloperons), fractions))


def _count_cluster_reannotations(operons: List[genes.Operon],
                                 reannotated_operons: Dict[str, genes.Operon]) -> ReannotationCounts:
    """
    After clustering operons by motif, the next question we'd often like to ask is how good
    the annotations are. This function takes the operons from one cluster, and operons
    that have been re-BLASTed against a more general database, and determines how often
    the annotations from the first round of BLASTing were reannotated in a parcticular way
    during the second round.

    For example, in a cluster with 130 operons, a particular region of DNA is annotated as cas9 using a curated
    database, and after BLASTing against nr, we find:

    - 107 S. pyogenes Cas9
    - 13 S. pyogenes DNA methyltransferase
    - 8 Unknown hypothetical protein
    - 2 Chloride intracellular channel protein 1

    While a small number of annotations look like other proteins, they're probably just noise, and we may
    tentatively conclude that this is a bonafide Cas9.

    After calculating the counts, this function then converts them to rough fractions:
    - 82 S. pyogenes Cas9
    - 10 S. pyogenes DNA methyltransferase
    - 6 Unknown hypothetical protein
    - 1 Chloride intracellular channel protein 1

    Partial overlaps with fraction greater than min_overlap_threshold coverage of the original operon
    will be included in the counts.

    """
    cluster_reannotations = defaultdict(dict)

    for operon in operons:
        reannotated_operon = reannotated_operons.get(operon.contig)
        if reannotated_operon is None:
            # We don't have a reannotation for this particular operon
            # This is common if we only re-BLAST a subset of our data
            # for efficiency's sake
            continue

        # Count the reannotations for a single pair of operons and merge those counts into the 
        # running total for the entire cluster
        update = _count_reannotations(operon, reannotated_operon)
        for feature_name, reannotation_data in update.items():
            for reannotated_feature_name, count in reannotation_data.items():
                current_count = cluster_reannotations[feature_name].get(reannotated_feature_name, 0)
                cluster_reannotations[feature_name][reannotated_feature_name] = count + current_count

            # If we get no reannotations for feature_name, we still need to add feature_name to the
            # cluster_reannotations dictionary, so that we can see when features are NEVER reannotated
            if not cluster_reannotations.get(feature_name):
                cluster_reannotations[feature_name] = defaultdict(int)
    return cluster_reannotations


def _count_reannotations(operon: genes.Operon,
                        reannotated_operon: genes.Operon):
    # To disambiguate features that occur more than once in each operon, we keep track of how many we have
    # of each. This way, we can refer to "transposase" and "transposase-2" as separate entities that
    # might BLAST to two completely different proteins.
    reannotations = {}
    feature_counts = defaultdict(int)
    for feature in operon:
        # determine which feature we're referring to in case more than one has the same name
        # if there is more than one, figure out what the number we need to append to the name should be
        count = feature_counts.get(feature.name)
        if not count:
            feature_name = (feature.name, 1)
            reannotations[feature_name] = defaultdict(int)
            feature_counts[feature.name] = 1
        else:
            feature_counts[feature.name] += 1
            feature_name = (feature.name, count + 1)
            if feature_name not in reannotations:
                reannotations[feature_name] = defaultdict(int)

        for reannotated_feature in reannotated_operon:
            if feature.start <= reannotated_feature.start and feature.end >= reannotated_feature.end:
                # reannotated_feature is completely overlapped by feature. It's possible
                # that there are multiple reannotated hits that map to this feature.
                # if that is the case, we could get >100% identities
                reannotations[feature_name][reannotated_feature.name] += 1
            elif reannotated_feature.start <= feature.start and reannotated_feature.end >= feature.end:
                # feature is completely overlapped by reannotated_feature
                reannotations[feature_name][reannotated_feature.name] += 1
    return reannotations


def _convert_reannotation_counts_to_fractions(reannotations: ReannotationCounts) -> ReannotationFractions:
    """ Normalizes raw counts to fractions of all reannotations. """
    fractional_reannotations = {}
    for feature_name, count_data in reannotations.items():
        fractional_reannotations[feature_name] = {}
        if not count_data:
            continue
        else:
            total = sum(count_data.values())
            for other_name, count in sorted(count_data.items(), key=lambda x: -x[1]):
                fraction = count / total
                fractional_reannotations[feature_name][other_name] = fraction
    return fractional_reannotations


def _format_reannotation_summary(contig_label: str,
                                 num_operons: int,
                                 fractional_reannotations: ReannotationCounts,
                                 show_top: Optional[int] = 3) -> str:
    """ Converts summary data into a formatted string that can be printed. """
    output = [f"{num_operons}-{contig_label}"]
    for (feature_name, feature_count), reannotated_feature_data in fractional_reannotations.items():
        formatted_feature_name = feature_name if feature_count == 1 else f"{feature_name}-{feature_count}"
        output.append(f"  {formatted_feature_name}:")
        limit = len(reannotated_feature_data) if show_top is None else show_top
        for _, (reannotated_feature_name, fraction) in zip(range(limit), reannotated_feature_data.items()):
            percentage = 100 * fraction
            output.append(f"    {percentage :4.0f}  {reannotated_feature_name}")
    return "\n".join(output)


def _prepare_operons_for_counting(operons: List[genes.Operon], compare_to_operons: List[genes.Operon]) \
        -> Tuple[DefaultDict[Tuple[str, ...], List[genes.Operon]], Dict[str, genes.Operon]]:
    """ Takes two lists of operons and reformats them so that we can perform our reannotation analysis. """
    clusters = analyze.cluster_operons_by_feature_order(operons)
    reannotated_operons = {operon.contig: operon for operon in compare_to_operons}
    return clusters, reannotated_operons


def _has_at_least_one_feature_with_fractional_matches(summary: ReannotationFractions,
                                                     feature_name: str,
                                                     regex_pattern: str,
                                                     min_matching_threshold: float):
    """
    Determines if the cluster has a feature that is reannotated to something that
    matches regex_pattern more frequently than min_matching_threshold.
    """
    assert 0 < min_matching_threshold <= 1.0
    rx = re.compile(regex_pattern, re.IGNORECASE)
    for (feature_name, _), fractions in summary.items():
        total = 0
        for other_feature_name, fraction in fractions.items():
            if rx.search(other_feature_name):
                total += fraction
        if total > min_matching_threshold:
            return True
    return False



""" Parses piler-cr output to extract the sequences of every spacer. """

from typing import List
from collections import namedtuple
from Bio.Seq import Seq


RepeatSpacer = namedtuple('RepeatSpacer', ['position', 'repeat_len', 'spacer_len', 'sequence'])
BrokenSpacer = namedtuple('BrokenSpacer', ['position', 'repeat_len', 'sequence'])


def parse_pilercr_output(text: str, start: int, end: int):
    """ Takes the text of :program:`pilercr` raw output, and a start and end coordinate for an operon's
    neighborhood, and extracts all spacers in that region. """
    if text is None:
        # There were no spacers in this contig
        return []
    text = text.split("\n")
    text = _find_first_entry(text)
    if text is None:
        # There were no spacers in this contig
        return []
    results = []
    while True:
        repeat_spacers, text = _parse_entry(text)
        valid_spacers = []
        for rs in repeat_spacers:
            if start <= rs.position <= end:
                valid_spacers.append(rs)
        if valid_spacers:
            results.append(valid_spacers)
        text = _find_next_entry(text)
        if not text:
            break
    return results


def _find_next_entry(text: List[str]):
    for n, line in enumerate(text):
        if line.startswith(">"):
            return text[n:]
        elif line.startswith("SUMMARY"):
            return False
    return False


def _find_first_entry(text: List[str]):
    for n, line in enumerate(text):
        if line.startswith(">"):
            return text[n:]


def _parse_entry(text: List[str]):
    repeat_spacers = []
    assert text[0].startswith(">")
    for n, line in enumerate(text[4:]):
        line = line.strip()
        if line.startswith("="):
            return repeat_spacers, text[n:]
        else:
            data = line.split()
            if len(data) == 7:
                position, repeat_len, _, spacer_len, _, _, spacer_seq = data
                if "N" in spacer_seq:
                    # don't allow spacers with any uncertainty
                    continue
                repeat_spacers.append(RepeatSpacer(int(position), int(repeat_len), int(spacer_len), Seq(spacer_seq)))
            elif len(data) == 6:
                position, repeat_len, _, _, _, broken_spacer_seq = data
                if "N" in broken_spacer_seq:
                    # don't allow spacers with any uncertainty
                    continue
                repeat_spacers.append(BrokenSpacer(int(position), int(repeat_len), Seq(broken_spacer_seq)))



from collections import defaultdict
import itertools
import math
import os
import re
import sys
import random
import string
from typing import Tuple, Dict, IO, List, Optional, Iterable, Any
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from dna_features_viewer import GraphicFeature, GraphicRecord
from operon_analyzer.genes import Operon
from operon_analyzer.parse import assemble_operons, read_pipeline_output
from operon_analyzer import analyze


ContigDescriptor = Tuple[str, str, int, int]  # contig accession ID, contig filename, start coordinate, end coordinate


def build_image_filename(operon: Operon, directory: str = None) -> str:
    """ Generates a name (and optionally a path) for the PNG we are going to produce. """
    filename = "{contig}-{start}-{end}.png".format(
                contig=operon.contig,
                start=operon.start,
                end=operon.end)
    if directory:
        return os.path.join(directory, filename)
    return filename


def calculate_adjusted_operon_bounds(operon: Operon, include_ignored: bool = True) -> Tuple[int, int]:
    """
    Calculates the offset that should be removed from each Feature's position
    as well as the total length of the operon.
    """
    low, high = sys.maxsize, 0
    # do one pass to find the bounds of the features we're interested in
    for feature in operon.all_features:
        if feature.ignored_reasons and not include_ignored:
            continue
        low = min(low, feature.start)
        high = max(high, feature.end)
    assert low < high
    return low, high - low


def _get_feature_color(feature_name: str, feature_colors: Dict[str, Any]) -> Any:
    # Decides what color a feature should be. The keys in `feature_colors` are either strings
    # that should exactly match a feature name, or a regular expression. The values of the dictionary
    # should be anything that matplotlib interprets as a color.

    default_color = feature_colors.get("", "blue")
    # Since there was no exact match, either the key is a regular expression or the feature is one
    # the user doesn't care about, in which case we'll try to use the default
    for pattern, color in feature_colors.items():
        if re.search(pattern, feature_name, re.IGNORECASE):
            return color
    return default_color


def _find_colormap_bounds(operons: List[Operon], color_by_blast_statistic, other_operons: Optional[List[Operon]] = None) -> Tuple[float, float]:
    if other_operons is None:
        other_operons = []
    lower, upper = None, None
    if color_by_blast_statistic is not None:
        lower = sys.maxsize
        upper = -sys.maxsize
        for operon in itertools.chain(operons, other_operons):
            for feature in operon:
                if feature.strand is None:
                    # This is a CRISPR array
                    continue
                value = getattr(feature, color_by_blast_statistic)
                lower = min(value, lower)
                upper = max(value, upper)
    return lower, upper


def plot_operons(operons: List[Operon],
                 output_directory: str,
                 plot_ignored: bool = True,
                 color_by_blast_statistic: Optional[str] = None,
                 feature_colors: Optional[dict] = {},
                 nucl_per_line: Optional[int] = None,
                 show_accession: bool = False,
                 show_description: bool = False):
    """ 
    Takes :class:`operon_analyzer.genes.Operon` objects and saves plots of them to disk. 

    Args:
        operons (list): Operons to be plotted.
        output_directory (str): Path to the directory to save operon plots to.
        plot_ignored (bool, optional): Toggles plotting of features that were marked as ignorable 
            by :class:`operon_analyzer.rules.FilterSet` .
        color_by_blast_statistic (str, optional): Map an alignment quality statistic using the virdis
            color scale. For a list of alignment statistics captured by Opfi, see :ref:`opfi-output-format` .
        feature_colors (dict, optional): If a labeled database was used during candidate identification,
            features can be colored accordingly using "label": "feature-color" pairs. For more information
            about labeling sequence databases, see :ref:`labeling-sequences` .
        nucl_per_line (int, optional): Length (in base pairs) to wrap gene diagrams on. 
        show_accession (bool, optional): Show the accession number of the best hit for each plotted feature.
        show_description (bool, optional): Show the description of the best hit for each plotted feature.
    """
    lower, upper = _find_colormap_bounds(operons, color_by_blast_statistic)
    for operon in operons:
        out_filename = build_image_filename(operon, output_directory)
        fig = create_operon_figure(operon, plot_ignored, feature_colors, color_by_blast_statistic=color_by_blast_statistic, colormin=lower, colormax=upper, nucl_per_line=nucl_per_line, show_accession=show_accession, show_description=show_description)
        if fig is None:
            continue
        save_operon_figure(fig, out_filename)


def _calculate_paired_figure_dimensions(operon: Operon, other: Operon, operon_length: int, plot_ignored: bool):
    """ Determines the figure height and width needed to make a stacked operon figure.
    This is essentially a bunch of magic heuristics that seem to come up with values that
    minimize whitespace and avoid overlapping feature labels. """
    height_top = min(6, int(max(2, math.sqrt(len(operon)))))
    height_bottom = min(6, int(max(2, math.sqrt(len(other)))))
    figure_width = max(int(operon_length/1000), 8)
    return height_top, height_bottom, figure_width


def plot_operon_pairs(operons: List[Operon], other_operons: List[Operon], output_directory: str,
                      color_by_blast_statistic: Optional[str] = None,
                      plot_ignored: bool = False, feature_colors: Optional[dict] = {}):
    """ 
    Takes two lists of presumably related Operons, pairs them up such that the pairs overlap the same genomic region,
    and plots one on top of the other. This allows side-by-side comparison of two different pipeline runs, so that you can, for example,
    run your regular pipeline, then re-BLAST with a more general protein database like nr, and easily see how the annotations differ. 

    Args:
        operons (list): Operons to be plotted.
        other_operons (list): Related operons to be plotted for comparison.
        output_directory (str): Path to the directory to save operon plots to.
        plot_ignored (bool, optional): Toggles plotting of features that were marked as ignorable 
            by :class:`operon_analyzer.rules.FilterSet` .
        color_by_blast_statistic (str, optional): Map an alignment quality statistic using the virdis
            color scale. For a list of alignment statistics captured by Opfi, see :ref:`opfi-output-format` .
        feature_colors (dict, optional): If a labeled database was used during candidate identification,
            features can be colored accordingly using "label": "feature-color" pairs. For more information
            about labeling sequence databases, see :ref:`labeling-sequences` .
    """
    lower, upper = _find_colormap_bounds(operons, color_by_blast_statistic, other_operons=other_operons)
    for operon, other in make_operon_pairs(operons, other_operons):
        out_filename = build_image_filename(operon, output_directory)
        plot_operon_pair(operon, other, lower, upper, out_filename, color_by_blast_statistic, plot_ignored, feature_colors)


def plot_operon_pair(operon: Operon,
                     other: Operon,
                     lower: float,
                     upper: float,
                     out_filename: str,
                     color_by_blast_statistic: Optional[str] = None,
                     plot_ignored: bool = False,
                     feature_colors: Optional[dict] = {}):
    # Calculate the figure size and the range of coordinates in the contig that we will plot
    lower_coordinates_bound, operon_length = calculate_adjusted_operon_bounds(operon, plot_ignored)
    upper_coordinates_bound = lower_coordinates_bound + operon_length
    height_top, height_bottom, figure_width = _calculate_paired_figure_dimensions(operon, other, operon_length, plot_ignored)

    # We create a Figure and two Axes and make DNA Features Viewer
    # use them so that we can stay in control of the dimensions
    fig, (ax1, ax2) = plt.subplots(nrows=2,
                                   ncols=1,
                                   figsize=(figure_width, height_top + height_bottom),
                                   gridspec_kw={"height_ratios": (height_top, height_bottom)})

    # Plot the top operon
    create_operon_figure(operon,
                         plot_ignored,
                         feature_colors,
                         color_by_blast_statistic=color_by_blast_statistic,
                         colormin=lower,
                         colormax=upper,
                         existing_ax=ax1,
                         figure_height=height_top, show_description=True)

    # Plot the bottom operon. We set bounds here so that it exactly
    # matches the location in the contig of the top operon
    create_operon_figure(other,
                         plot_ignored,
                         feature_colors,
                         color_by_blast_statistic=color_by_blast_statistic,
                         colormin=lower,
                         colormax=upper,
                         bounds=(lower_coordinates_bound, upper_coordinates_bound),
                         existing_ax=ax2,
                         figure_height=height_bottom, show_description=True)

    # Save the figure to disk
    save_pair_figure(fig, out_filename)


def make_operon_pairs(operons: List[Operon], other: List[Operon]) -> List[Tuple[Operon, Operon]]:
    """ Takes two lists of operons and tries to find matching pairs that overlaps the same genomic region.
    The overlaps can be partial, or the span of one can be a subset of the other. The idea here is that
    if we run our pipeline on the same sequencing data with two different databases, the neighborhoods
    might not line up exactly, but we want to do direct comparisons of the same region (for example, you might
    run your seed/BLAST/filter database, and then want to re-BLAST with Swissprot or nr to refine the 
    matches).

    The regions covered by operons in `operons` are used as the reference point. """
    other_lookup = defaultdict(list)
    for operon in other:
        other_lookup[operon.contig].append(operon)

    pairs = []
    for operon in operons:
        candidates = other_lookup.get(operon.contig)
        if not candidates:
            continue
        best_overlap = 0
        best_candidate = None
        for candidate in candidates:
            overlap = _calculate_operon_overlap(operon, candidate)
            if overlap is None:
                continue
            if overlap >= 1.0:
                best_candidate = candidate
                best_overlap = overlap
                break
            if overlap > best_overlap:
                best_candidate = candidate
                best_overlap = overlap
        if best_candidate is not None:
            pairs.append((operon, best_candidate))
    return pairs


def _calculate_operon_overlap(operon: Operon, other_operon: Operon) -> Optional[float]:
    """ Calculates the fraction of an operon that overlaps with another operon. """
    operon_start, operon_end = min(operon.start, operon.end), max(operon.start, operon.end)
    other_operon_start, other_operon_end = min(other_operon.start, other_operon.end), max(other_operon.start, other_operon.end)

    if operon_start > other_operon_end or other_operon_start > operon_end:
        # these operons don't overlap at all
        return None
    # Find the lower of the two ends
    end = min(operon_end, other_operon_end)
    # Find the higher of the two starts
    start = max(operon_start, other_operon_start)
    # Determine how much overlap there is
    operon_length = operon_end - operon_start + 1
    return (end - start + 1) / operon_length


def create_operon_figure(operon: Operon,
                         plot_ignored: bool,
                         feature_colors: Optional[dict] = {},
                         bounds: Optional[Tuple[int, int]] = None,
                         existing_ax: Optional[Axes] = None,
                         color_by_blast_statistic: Optional[str] = None,
                         colormin: Optional[float] = None,
                         colormax: Optional[float] = None,
                         figure_height: Optional[int] = None,
                         nucl_per_line: Optional[int] = None,
                         show_accession: bool = False,
                         show_description: bool = False):
    """ Plots all the Features in an Operon. """
    if not plot_ignored and len(operon) == 0:
        return None

    if colormin is not None and colormax is not None:
        norm = matplotlib.colors.LogNorm(vmin=colormin + sys.float_info.min, vmax=colormax)
        cmap = matplotlib.cm.get_cmap('viridis_r')

    if not bounds:
        offset, operon_length = calculate_adjusted_operon_bounds(operon, plot_ignored)
    else:
        offset = bounds[0]
        operon_length = bounds[1] - bounds[0]

    graphic_features = []
    for feature in operon.all_features:
        if feature.ignored_reasons and not plot_ignored:
            continue

        if colormin is not None and colormax is not None and not feature_colors:
            value = getattr(feature, color_by_blast_statistic)
            if value is None:
                color = 'gray'
            elif value == 0:
                color = cmap(0)
            else:
                normed_value = norm(value)
                color = cmap(normed_value)
        else:
            color = _get_feature_color(feature.name, feature_colors)
        if bounds and (not bounds[0] <= feature.start or not bounds[1] >= feature.end):
            continue
        # we alter the name of CRISPR arrays to add the number of repeats
        # this is done here and not earlier in the pipeline so that it doesn't
        # affect any rules that need to match on the name
        name = feature.name
        if show_accession:
            name = f"{name} {feature.accession}"
        if show_description:
            name = f"{name} {feature.description}"
        strand = feature.strand
        if feature.name == "CRISPR array":
            copies, repeat, spacer = feature.description.split(",")
            _, count = copies.split()
            name = f"CRISPR array ({count})"
            strand = None  # don't let array have a directional arrow
        label = name if not feature.ignored_reasons else f"{name} (ignored)"
        graphic_feature = GraphicFeature(start=feature.start - offset,
                                         strand=strand,
                                         end=feature.end - offset,
                                         label=label,
                                         color=color)
        graphic_features.append(graphic_feature)
    record = GraphicRecord(sequence_length=operon_length,
                           features=graphic_features)

    if nucl_per_line is not None:
        fig, _ = record.plot_on_multiple_lines(nucl_per_line=nucl_per_line, max_label_length=255)
    else:
        figure_width = max(int(operon_length/900), 5)
        ax, _ = record.plot(figure_width=figure_width, figure_height=figure_height, ax=existing_ax, max_label_length=255)
        fig = ax.figure
    return fig


def save_operon_figure(fig, out_filename: str):
    """ Writes the operon figure to disk. """
    fig.savefig(out_filename, bbox_inches='tight')
    plt.close()


def save_pair_figure(fig, out_filename: str):
    """ Writes the operon pair figure to disk. """
    fig.savefig(out_filename, bbox_inches='tight')
    plt.close()


def build_operon_dictionary(f: IO[str]) -> Dict[Tuple[str, int, int], Operon]:
    """ Builds a dictionary of Operons since we need to be able to randomly access them. """
    operons = {}
    lines = read_pipeline_output(f)
    for operon in assemble_operons(lines):
        operons[(operon.contig, operon.contig_filename, operon.start, operon.end)] = operon
    return operons


def _load_passing_contigs(handle: IO):
    good_contigs = set()
    for contig, contig_filename, start, end, result in analyze.load_analyzed_operons(handle):
        if result[0] != 'pass':
            continue
        good_contigs.add((contig, contig_filename, start, end))
    return good_contigs


def make_clustered_stacked_operon_plots(operons: Iterable[Operon],
                                        other_operons: Iterable[Operon],
                                        image_directory: str,
                                        min_count: int = 10,
                                        plot_ignored: bool = False,
                                        color_by_blast_statistic: Optional[str] = None,
                                        feature_colors: Optional[dict] = None
                                        ):
    """
    Clusters operons and plots them on top of a reannotated version of the same operon. This allows the user to BLAST
    some set of data with a curated database, then re-BLAST it against a more general database, and compare the two
    directly in a cluster-specific manner.

    If a :class:`operon_analyzer.rules.FilterSet` was used during analysis, that same set should be evaluated on each operon before passing it
    into this function.

    Args:
        operons (iterable): The operons of interest.
        other_operons (iterable): Reannotated operons.
        image_directory (str): The directory where all subdirectories will be created. Will be created if it does not exist.
        min_count (int, optional): Groups must have at least this many systems in order to be plotted. Default is 10.
        plot_ignored (bool, optional): Toggles plotting of features that were marked as ignorable 
            by :class:`operon_analyzer.rules.FilterSet` .
        color_by_blast_statistic (str, optional): Map an alignment quality statistic using the virdis
            color scale. For a list of alignment statistics captured by Opfi, see :ref:`opfi-output-format` .
        feature_colors (dict, optional): If a labeled database was used during candidate identification,
            features can be colored accordingly using "label": "feature-color" pairs. For more information
            about labeling sequence databases, see :ref:`labeling-sequences` .
        nucl_per_line (int, optional): Length (in base pairs) to wrap gene diagrams on. 
    """
    if not os.path.exists(image_directory):
        os.makedirs(image_directory)

    clustered_operons = analyze.cluster_operons_by_feature_order(operons)
    _plot_clustered_stacked_operons(clustered_operons, other_operons, image_directory, plot_ignored=plot_ignored, color_by_blast_statistic=color_by_blast_statistic, feature_colors=feature_colors)


def make_clustered_operon_plots(analysis_csv: str,
                                operons: Iterable[Operon],
                                image_directory: str,
                                min_count: int = 10,
                                diff_against_csv: Optional[str] = None,
                                plot_ignored: bool = False,
                                color_by_blast_statistic: Optional[str] = None,
                                feature_colors: Optional[dict] = None,
                                nucl_per_line: Optional[int] = None
                                ):
    """ 
    Clusters operons by the order of their features and plots them in separate directories,
    adding the number of systems to the directory name. Only systems that passed the rules specified as a
    :class:`operon_analyzer.rules.RuleSet` object will be eligible to be plotted.

    If a :class:`operon_analyzer.rules.FilterSet` was used during analysis, that same FilterSet should be evaluated 
    on each operon before passing it into this function.

    Args:
        analysis_csv (str): Path to the CSV file created by :func:`operon_analyzer.analyze.analyze` .
        operons (iterable): The operons of interest.
        image_directory (str): The directory where all subdirectories will be created. Will be created if it does not exist.
        min_count (int, optional): Groups must have at least this many systems in order to be plotted. Default is 10.
        diff_against_csv (str): Path to a CSV file created by operon analyzer. Any clusters in this file 
            will be skipped when clustering operons from analysis_csv. The point of 
            this is to see only new systems when making slight alterations to rules.
        plot_ignored (bool, optional): Toggles plotting of features that were marked as ignorable 
            by :class:`operon_analyzer.rules.FilterSet` .
        color_by_blast_statistic (str, optional): Map an alignment quality statistic using the virdis
            color scale. For a list of alignment statistics captured by Opfi, see :ref:`opfi-output-format` .
        feature_colors (dict, optional): If a labeled database was used during candidate identification,
            features can be colored accordingly using "label": "feature-color" pairs. For more information
            about labeling sequence databases, see :ref:`labeling-sequences` .
    """
    with open(analysis_csv) as f:
        passing_contigs = _load_passing_contigs(f)

    diff_contigs = set()
    if diff_against_csv is not None:
        with open(diff_against_csv) as f:
            diff_contigs = _load_passing_contigs(f)

    if not os.path.exists(image_directory):
        os.mkdir(image_directory)

    plottable_operons = []
    diff_operons = []

    for operon in operons:
        key = operon.contig, operon.contig_filename, operon.start, operon.end
        if key in passing_contigs:
            plottable_operons.append(operon)
        if key in diff_contigs:
            diff_operons.append(operon)

    all_clustered_operons = analyze.cluster_operons_by_feature_order(plottable_operons)
    diff_clustered_operon_motifs = set(analyze.cluster_operons_by_feature_order(diff_operons).keys())
    plottable_operons = {k: ops for k, ops in all_clustered_operons.items()
                         if k not in diff_clustered_operon_motifs and len(ops) >= min_count}
    _plot_clustered_operons(plottable_operons, image_directory, plot_ignored, color_by_blast_statistic=color_by_blast_statistic, feature_colors=feature_colors, nucl_per_line=nucl_per_line)


def _plot_clustered_stacked_operons(clustered_operons: Dict[str, List[Operon]],
                                    other_operons: List[Operon],
                                    image_directory: str,
                                    plot_ignored: bool,
                                    color_by_blast_statistic: Optional[str] = None,
                                    feature_colors: Optional[dict] = None):
    for motif_items, ops in clustered_operons.items():
        motif_name = '-'.join(motif_items)
        motif_directory = _make_motif_directory_name(motif_name, len(ops), image_directory)
        os.makedirs(motif_directory)
        plot_operon_pairs(ops, other_operons, motif_directory, plot_ignored=plot_ignored, color_by_blast_statistic=color_by_blast_statistic, feature_colors=feature_colors)


def _plot_clustered_operons(clustered_operons: Dict[str, List[Operon]], image_dir: str, plot_ignored: bool, color_by_blast_statistic: Optional[str] = None, feature_colors: Optional[dict] = None, nucl_per_line: Optional[int] = None):
    """ Plots contigs, placing them in directories named by the count and motif of the operons.
    For example, if there are 527 operons with Cas9, glmS, a CRISPR array, and Cas1 (in that order or exactly reversed),
    the directory name will be 527-cas9-glms-array-cas1. """
    for motif_items, operons in clustered_operons.items():
        motif_name = '-'.join(motif_items)
        motif_directory = _make_motif_directory_name(motif_name, len(operons), image_dir)
        os.mkdir(motif_directory)
        plot_operons(operons, motif_directory, plot_ignored=plot_ignored, color_by_blast_statistic=color_by_blast_statistic, feature_colors=feature_colors, nucl_per_line=nucl_per_line)


def _make_motif_directory_name(motif_name: str, num_operons: int, image_dir: str) -> str:
    """ Makes a string to use as a directory name for some set of operons with the same motif """
    motif_name = motif_name.replace("CRISPR array", "array").replace(" ", "_")
    motif_name = f"{num_operons}-{motif_name}"
    # get the filename character limit for the file system we are trying to write to
    max_allowed_chars = os.statvfs(image_dir).f_namemax
    if len(motif_name) > max_allowed_chars:
        motif_name = motif_name[:max_allowed_chars - 10] + "-" + ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
    return os.path.join(image_dir, motif_name)



from typing import List, Union, Optional, Tuple
from collections import namedtuple
import tempfile
import subprocess
import statistics
import json
import multiprocessing
from Bio.Seq import Seq
from operon_analyzer import genes, piler_parse, load
import parasail
import re


GAP_OPEN_PENALTY = 8
GAP_EXTEND_PENALTY = 8

AlignmentResult = namedtuple('AlignmentResult', ['match_count', 'spacer_sequence', 'contig_sequence', 'contig_start', 'contig_end', 'spacer_alignment', 'contig_alignment', 'comp_string', 'strand', 'spacer_order', 'array_length'])
MAX_SPACER_LENGTH_BP = 40  # Assume spacers longer than this were incorrectly parsed and should be ignored
Spacer = Union[piler_parse.RepeatSpacer, piler_parse.BrokenSpacer]
Array = List[Spacer]

cigar_regex = re.compile(r"(\d+)=")
start_regex = re.compile(r'^(\d+)D')


def find_self_targeting_spacers(operons: List[genes.Operon], min_matching_fraction: float, num_processes: int = 32):
    """
    For each given Operon, this will determine if CRISPR spacers target a location in the operon's parent contig.
    Matches with more than ``min_matching_fraction`` homology will be added to the Operon as a Feature named "CRISPR target".
    """
    pool = multiprocessing.Pool(min(num_processes, len(operons)))
    results = []
    for operon in operons:
        result = pool.apply_async(_align_operon_spacers, args=(operon, min_matching_fraction))
        results.append(result)
    pool.close()
    pool.join()
    for result in results:
        yield result.get()


def _align_operon_spacers(operon: genes.Operon, min_matching_fraction: float):
    """
    Aligns every spacer in an Operon to its parent contig. Any spacers with a target that has
    at least min_matching_fraction homology will be converted to Features with the name
    "CRISPR target" and added to the Operon object.
    """
    contig_sequence = load.load_sequence(operon)
    assert contig_sequence is not None, f"Operon's sequence file cannot be found: {operon.contig_filename}."
    spacers = _get_operon_spacers(operon.start, operon.end, contig_sequence)
    if not spacers:
        return operon

    # Look up where the arrays are. If a spacer targets the array it comes from, we don't create a CRISPR target.
    # It's either an assembly artifact or a duplication of the array.
    arrays = [feature for feature in operon.all_features if feature.name == 'CRISPR array']

    for spacer, spacer_order, array_length in spacers:
        ar = _perform_local_pairwise_alignment(spacer, spacer_order, array_length, str(contig_sequence))
        if not ar:
            continue
        matching_fraction = ar.match_count / len(ar.spacer_sequence)
        if matching_fraction >= min_matching_fraction:
            feature = _build_feature_from_alignment(ar, spacer)

            # Make sure the spacer doesn't target its own array
            for array in arrays:
                if array.start <= feature.start <= array.end and array.start <= spacer.position <= array.end:
                    # The spacer and target are both in the same array
                    break
            else:
                # The target is somewhere outside of the array where its spacer is
                operon._features.append(feature)
    return operon


def _get_operon_spacers(operon_start: int, operon_end: int, contig_sequence: Seq):
    """
    Runs pilercr and extracts all the spacer sequences found within the operon.
    Also adds the position of the spacer in the array, and the total array length.
    """
    piler_data = _run_piler(contig_sequence)
    arrays = piler_parse.parse_pilercr_output(piler_data, operon_start, operon_end)
    fixed_arrays = _fix_arrays(arrays, contig_sequence)
    spacers = []
    for array in fixed_arrays:
        for n, spacer in enumerate(array):
            if len(spacer.sequence) > MAX_SPACER_LENGTH_BP:
                continue
            spacers.append((spacer, n, len(array)))
    return spacers


def _perform_local_pairwise_alignment(spacer: piler_parse.RepeatSpacer,
                                      spacer_order: int,
                                      array_length: int,
                                      contig: Seq) -> Optional[AlignmentResult]:
    """
    Aligns a spacer and its reverse complement to the given contig and returns the better of the two matches.
    The spacer sequence itself is removed from the contig so that we don't just return the spacer itself.
    """
    best_score = 0
    best_strand = None
    assert str(spacer.sequence) in contig
    censored_contig = _build_censored_contig(spacer, contig)

    spacer_seqs = spacer.sequence, spacer.sequence.reverse_complement()

    for n, spacer_seq in enumerate(spacer_seqs):
        results = _align(str(spacer_seq), censored_contig)
        score = results[0]
        if score > best_score:
            best_score = score
            best_results = results
            best_strand = n

    _, match_count, contig_target, contig_start, contig_end, contig_alignment, spacer_alignment, comp_string = best_results
    strand = [1, -1][best_strand]

    alignment_result = AlignmentResult(match_count,
                                       str(spacer_seqs[best_strand]),
                                       contig_target,
                                       contig_start,
                                       contig_end,
                                       spacer_alignment,
                                       contig_alignment,
                                       comp_string,
                                       strand,
                                       spacer_order,
                                       array_length)

    # We shouldn't have more matching base pairs than base pairs
    assert alignment_result.match_count <= len(alignment_result.spacer_sequence), str(alignment_result)
    assert alignment_result.match_count <= len(alignment_result.contig_sequence), str(alignment_result)

    if "N" in alignment_result.contig_sequence:
        # don't permit contigs with any uncertainty
        return None
    return alignment_result


def _build_censored_contig(spacer: piler_parse.RepeatSpacer, contig: Seq) -> str:
    """
    Replace the spacer sequence itself in the contig with repeating "N"s so that
    the local alignment doesn't just find the spacer itself.
    """
    spacer_start = spacer.position + spacer.repeat_len
    spacer_end = spacer.position + spacer.repeat_len + spacer.spacer_len
    censored = str(contig[:spacer_start]) + "N"*spacer.spacer_len + str(contig[spacer_end:])
    assert len(censored) == len(contig), f"{len(contig)} {len(censored)}"
    return censored


def _align(spacer: str, contig: str) -> Tuple[float, int, str, int, int, str, str, str]:
    """ Run the local pairwise alignment of two strings and return alignment data. """
    # perform the alignment
    result = parasail.sw_trace(spacer, contig, GAP_OPEN_PENALTY, GAP_EXTEND_PENALTY, parasail.nuc44)

    # extract pertinent data from the alignment result
    cigar_text = result.cigar.decode.decode("utf8")
    match_count = _count_cigar_matches(cigar_text)
    contig_target_sequence = result.traceback.ref.replace("-", "")
    contig_end = result.end_ref
    contig_start = contig_end - len(contig_target_sequence) + 1
    return result.score, match_count, contig_target_sequence, contig_start, contig_end, result.traceback.ref, result.traceback.query, result.traceback.comp


def _build_feature_from_alignment(ar: AlignmentResult, spacer: piler_parse.RepeatSpacer) -> genes.Feature:
    """ Alignment results are converted to Features so that we can plot them easily. """
    description = json.dumps({"match_count": ar.match_count,
                              "matching_fraction": (ar.match_count / len(str(ar.spacer_sequence))),
                              "spacer_alignment": ar.spacer_alignment,
                              "contig_alignment": ar.contig_alignment,
                              "contig_start": ar.contig_start,
                              "contig_end": ar.contig_end,
                              "spacer_sequence": str(ar.spacer_sequence),
                              "contig_sequence": str(ar.contig_sequence),
                              "comp_string": ar.comp_string,
                              "spacer_order": ar.spacer_order,
                              "spacer_position": spacer.position,
                              "array_length": ar.array_length})
    feature = genes.Feature('CRISPR target',
                            (ar.contig_start, ar.contig_end),
                            '',
                            ar.strand,
                            '',
                            None,
                            description,
                            '',
                            None)
    return feature


def _count_cigar_matches(string: str) -> int:
    """ parasail provides a CIGAR string to encode the alignment. We parse this to determine the number of exact matches. """
    matches = cigar_regex.findall(string)
    return sum([int(match) for match in matches])


def _run_piler(sequence: str) -> str:
    """ Runs pilercr on the given Seq sequence and returns the raw file output """
    with tempfile.NamedTemporaryFile('w', dir='/tmp') as contig_file, tempfile.NamedTemporaryFile('w', dir='/tmp') as pilercr_file:
        contig_file.write(f">sequence\n{sequence}")
        command = ["pilercr", "-in", contig_file.name,
                              "-out", pilercr_file.name,
                              "-minarray", "2",
                              "-quiet",
                              "-noinfo"]
        # Piler crashes on some inputs. We pipe stderr to suppress the error messages.
        # We may be losing results when this occurs, but it's pretty rare and there's
        # nothing we can do about it either way.
        result = subprocess.call(command, stderr=subprocess.DEVNULL)
        if result != 0:
            return None
        with open(pilercr_file.name) as f:
            return f.read()


def _fix_arrays(arrays: List[Array], contig: Seq) -> List[Array]:
    """
    Ensures that the final spacer in each array is a reasonable length. For some reason, they
    are occasionally substantially shorter than all preceeding spacers in the array.
    """
    fixed_arrays = []
    for array in arrays:
        median_length = _find_median_spacer_length(array)
        if not median_length:
            continue
        fixed_array = _fix_array(array, median_length, contig)
        fixed_arrays.append(fixed_array)
    return fixed_arrays


def _find_median_spacer_length(array: List[Spacer]) -> Optional[int]:
    good_spacers = [rs.spacer_len for rs in array if type(rs) == piler_parse.RepeatSpacer]
    if not good_spacers:
        return None
    return int(statistics.median(good_spacers))


def _fix_array(array: List[Spacer], median_length: int, sequence: Seq) -> List[piler_parse.RepeatSpacer]:
    """
    Pilercr sometimes reports that the last spacer in an array is much shorter than the rest.
    We assume that it should at least have the median length of the rest of the spacers and extend
    it to that length.
    """
    fixed_array = []
    for rs in array:
        # Figure out what the spacer sequence is
        if type(rs) is piler_parse.BrokenSpacer:
            fixed_rs = _fix_broken_spacer(rs, median_length, sequence)
            if fixed_rs:
                fixed_array.append(fixed_rs)
        else:
            fixed_array.append(rs)
    return fixed_array


def _fix_broken_spacer(rs: piler_parse.BrokenSpacer, median_length: int, sequence: Seq) -> Optional[piler_parse.RepeatSpacer]:
    spacer = str(sequence)[rs.position + rs.repeat_len: rs.position + rs.repeat_len + median_length]
    if len(spacer) > MAX_SPACER_LENGTH_BP:
        return None
    return piler_parse.RepeatSpacer(rs.position, rs.repeat_len, len(spacer), Seq(spacer))



import re
from typing import Callable, Optional, List
from operon_analyzer.genes import Feature, Operon
import more_itertools


class SerializableFunction(object):
    """ A base class for functions that we need to be able to serialize. Do not instantiate this directly. """

    def __init__(self, name: str, function: Callable, *args, custom_repr: Optional[str] = None):
        self._name = name
        self._function = function
        self._args = args
        self._custom_repr = custom_repr

    def __repr__(self) -> str:
        if self._custom_repr is not None:
            return self._custom_repr
        return "{name}:{args}".format(
                name=str(self._name),
                args="-".join(map(str, self._args)))


class Rule(SerializableFunction):
    """ Defines a requirement that elements of an operon must adhere to. """

    def evaluate(self, operon: Operon) -> bool:
        """ Determine if an operon adheres to this rule. """
        return self._function(operon, *self._args)


class Result(object):
    """
    Records which rules an operon passed and handles serialization of the data.
    Also makes it easy to run follow-up queries.

    """

    def __init__(self, operon: Operon):
        self.operon = operon
        self._passing = []  # type: List[Rule]
        self._failing = []  # type: List[Rule]

    def add_passing(self, rule: Rule):
        """ Mark this rule as being one that the operon passed. """
        self._passing.append(rule)

    def add_failing(self, rule: Rule):
        """ Mark this rule as being one that the operon failed. """
        self._failing.append(rule)

    @property
    def is_passing(self) -> bool:
        """ Declares whether the given Operon adhered to all given Rules. """
        assert (self._passing or self._failing), "Result has no Rules!"
        return bool(self._passing) and not bool(self._failing)


class Filter(SerializableFunction):
    """
    A function that will be run on an Operon that marks Features as being ignorable
    for the purposes of evaluating RuleSets.
    """

    def run(self, operon: Operon):
        """ Mark Features in the Operon as ignored if they don't pass the filter. This
        will prevent them from being taken into account during Rule evaluation and (by
        default) during visualization. """
        return self._function(operon, str(self), *self._args)


class FilterSet(object):
    """
    Stores functions that take an Operon and mark individual Features as ignored
    in case we think they are not actually worth taking into account when evaluating rules.
    Features can be ignored for multiple reasons.
    """

    def __init__(self):
        self._filters = []

    def must_be_within_n_bp_of_anything(self, distance_bp: int):
        """ If a feature is very far away from anything it's probably not part of an operon. """
        self._filters.append(Filter('must-be-within-n-bp-of-anything', _must_be_within_n_bp_of_anything, distance_bp))
        return self

    def must_be_within_n_bp_of_feature(self, feature_name: str, distance_bp: int, regex: bool = False):
        """ There may be situations where two features always appear near each other in functional operons. """
        self._filters.append(Filter('must-be-within-n-bp-of-feature', _must_be_within_n_bp_of_feature, feature_name, distance_bp, regex))
        return self

    def pick_overlapping_features_by_bit_score(self, minimum_overlap_threshold: float):
        """ If two features overlap by more than ``minimum_overlap_threshold``, the one with the lower bit score is ignored. """
        self._filters.append(Filter('overlaps-%s',
                                    _pick_overlapping_features_by_bit_score,
                                    minimum_overlap_threshold))
        return self

    def custom(self, filt: 'Filter'):
        """ Add a rule with a user-defined function. """
        self._filters.append(filt)
        return self

    def evaluate(self, operon: Operon):
        """ 
        Run the filters on the operon and set Features that fail to meet the requirements to be ignored. 

        Args:
            operon: The :class:`operon_analyzer.genes.Operon` object whose features will be evaluated.
        """
        for filt in self._filters:
            filt.run(operon)


def _pick_overlapping_features_by_bit_score(operon: Operon, ignored_reason_message: str, minimum_overlap_threshold: float):
    """ If BLAST identified two genes at the same location, we want to determine which one we should
    actually consider the ORF to actually be. In these cases, we pick whichever one has the highest
    bit score. Since there are many small overlaps, the user must specify some arbitrary limit to
    define what an overlap is. """

    for feature in operon.all_genes:
        for other_feature in operon.all_genes:
            if feature is other_feature:
                # don't compare feature to itself
                continue
            overlap = _calculate_overlap(feature, other_feature)
            if overlap is None:
                # these features do not overlap
                continue
            if overlap >= minimum_overlap_threshold and feature.bit_score < other_feature.bit_score:
                feature.ignore(ignored_reason_message % other_feature.name)


def _calculate_overlap(feature: Feature, other_feature: Feature) -> Optional[float]:
    """ Calculates the fraction of a feature that overlaps with another feature. """
    if feature.start > other_feature.end or other_feature.start > feature.end:
        # these features don't overlap at all
        return None
    # Find the lower of the two ends
    end = min(feature.end, other_feature.end)
    # Find the higher of the two starts
    start = max(feature.start, other_feature.start)
    # Determine how much overlap there is
    feature_length = feature.end - feature.start + 1
    return (end - start + 1) / feature_length


def _must_be_within_n_bp_of_anything(operon: Operon, ignored_reason_message: str, distance_bp: int):
    for feature in operon.all_features:
        distances = _calculate_all_distances(operon, feature)
        if not distances:
            continue
        if min(distances) > distance_bp:
            feature.ignore(ignored_reason_message)


def _must_be_within_n_bp_of_feature(operon: Operon, ignored_reason_message: str, feature_name: str, distance_bp: int, regex: bool):
    reg = re.compile(feature_name, re.IGNORECASE)
    for feature in operon.all_features:
        if reg.match(feature.name):
            continue
        if not _max_distance(operon, feature_name, feature.name, distance_bp, False, regex):
            feature.ignore(ignored_reason_message)


class RuleSet(object):
    """ Creates, stores and evaluates :class:`operon_analyzer.rules.Rule` s that an operon must adhere to."""

    def __init__(self):
        self._rules = []

    def exclude(self, feature_name: str, regex: bool = False):
        """ Forbid the presence of a particular feature. """
        self._rules.append(Rule('exclude', _exclude, feature_name, regex))
        return self

    def require(self, feature_name: str, regex: bool = False):
        """ Require the presence of a particular feature. """
        self._rules.append(Rule('require', _require, feature_name, regex))
        return self

    def max_distance(self, feature1_name: str, feature2_name: str, distance_bp: int, closest_pair_only: bool = False, regex: bool = False):
        """
        The two given features must be no further than ``distance_bp`` base pairs
        apart. If there is more than one match, all possible pairs must meet the criteria,
        unless ``closest_pair_only`` is True in which case only the closets pair is considered.
        """
        self._rules.append(Rule('max-distance', _max_distance, feature1_name, feature2_name, distance_bp, closest_pair_only, regex))
        return self

    def at_least_n_bp_from_anything(self, feature_name: str, distance_bp: int, regex=False):
        """
        Requires that a feature be at least ``distance_bp`` base pairs away from any other feature.
        This is mostly useful for eliminating overlapping features.
        """
        self._rules.append(Rule('at-least-n-bp-from-anything', _at_least_n_bp_from_anything, feature_name, distance_bp, regex))
        return self

    def at_most_n_bp_from_anything(self, feature_name: str, distance_bp: int, regex: bool = False):
        """
        A given feature must be within ``distance_bp`` base pairs of another feature.
        Requires exactly one matching feature to be present.
        Returns False if the given feature is the only feature.
        """
        self._rules.append(Rule('at-most-n-bp-from-anything', _at_most_n_bp_from_anything, feature_name, distance_bp, regex))
        return self

    def same_orientation(self, exceptions: Optional[List[str]] = None):
        """
        All features in the operon must have the same orientation.
        """
        self._rules.append(Rule('same-orientation', _same_orientation, exceptions))
        return self

    def contains_any_set_of_features(self, sets: List[List[str]]):
        """
        Returns True if the operon contains features with all of the names
        in at least one of the lists. Useful for determining if an operon contains
        all of the essential genes for a particular system, for example.
        """
        serialized_sets = "|".join(["-".join(names) for names in sets])
        custom_repr = f'contains-any-set-of-features:{serialized_sets}'
        self._rules.append(Rule('contains-any-set-of-features',
                                _contains_any_set_of_features,
                                sets, custom_repr=custom_repr))
        return self

    def contains_exactly_one_of(self, feature1_name: str, feature2_name: str, regex: bool = False):
        """
        An exclusive-or of the presence of two features.
        That is, one of the features must be present and the other must not.
        """
        self._rules.append(Rule('contains-exactly-one-of',
                                _contains_exactly_one_of,
                                feature1_name,
                                feature2_name))
        return self

    def contains_at_least_n_features(self, feature_names: List[str], feature_count: int, count_multiple_copies: bool = False):
        """
        The operon must contain at least ``feature_count`` features in the list. By default, a
        matching feature that appears multiple times in the operon will only be counted once;
        to count multiple copies of the same feature, set ``count_multiple_copies`` to True.
        """
        serialized_list = "|".join(feature_names)
        custom_repr = f'contains-at-least-n-features:{serialized_list}-{feature_count}-{count_multiple_copies}'
        self._rules.append(Rule('contains_at_least_n_features',
                                _contains_at_least_n_features,
                                feature_names,
                                feature_count,
                                count_multiple_copies,
                                custom_repr=custom_repr))
        return self

    def contains_group(self, feature_names: List[str], max_gap_distance_bp: int, require_same_orientation: bool):
        """ The operon must contain a contiguous set of features (in any order) separated by no more than ``max_gap_distance_bp``.
        Optionally, the user may require that the features must all have the same orientation. """
        self._rules.append(Rule('contains_group',
                                _contains_group,
                                feature_names,
                                max_gap_distance_bp,
                                require_same_orientation))
        return self

    def maximum_size(self, feature_name: str, max_bp: int, all_matching_features_must_pass: bool = False, regex: bool = False):
        """ The operon must contain at least one feature with ``feature_name`` with a size (in base pairs) of ``max_bp`` or smaller.
        If ``all_matching_features_must_pass`` is True, every matching Feature must be at least ``max_bp`` long. """
        self._rules.append(Rule('maximum_size',
                                _maximum_size,
                                feature_name,
                                max_bp,
                                all_matching_features_must_pass,
                                regex))
        return self

    def minimum_size(self, feature_name: str, min_bp: int, all_matching_features_must_pass: bool = False, regex: bool = False):
        """ The operon must contain at least one feature with ``feature_name`` with a size (in base pairs) of ``min_bp`` or larger.
        If ``all_matching_features_must_pass`` is True, every matching Feature must be at least ``min_bp`` long. """
        self._rules.append(Rule('minimum_size',
                                _minimum_size,
                                feature_name,
                                min_bp,
                                all_matching_features_must_pass,
                                regex))
        return self

    def custom(self, rule: 'Rule'):
        """ Add a rule with a user-defined function. """
        self._rules.append(rule)
        return self

    def evaluate(self, operon: Operon) -> Result:
        """ 
        See if an operon adheres to all rules.

        Args:
            operon: The :class:`operon_analyzer.genes.Operon` object to evaluate.
        """
        assert self._rules, "RuleSet does not have any Rules"
        result = Result(operon)
        for rule in self._rules:
            if rule.evaluate(operon):
                result.add_passing(rule)
            else:
                result.add_failing(rule)
        return result

    def __repr__(self) -> str:
        return ",".join((str(rule) for rule in self._rules))


def _exclude(operon: Operon, feature_name: str, regex: str) -> bool:
    """ Returns false if a feature's name in the operon matches the given string. """
    return not _require(operon, feature_name, regex)


def _require(operon: Operon, feature_name: str, regex: bool) -> bool:
    """ Returns true if a feature's name in the operon matches the given string. """
    if regex:
        rx = re.compile(feature_name, re.IGNORECASE)
        return any([rx.match(name) for name in operon.feature_names])
    return feature_name.lower() in map(str.lower, operon.feature_names)


def _max_distance(operon: Operon, feature1_name: str, feature2_name: str, distance_bp: int, closest_pair_only: bool, regex: bool) -> bool:
    """ Returns whether two given Features are within distance_bp base pairs from each other.
    This must hold for all copies of the features match unless closest_pair_only is True,
    in which case we only consider the pair that is the shortest distance apart. """
    distances = []
    for f1 in operon.get(feature1_name, regex):
        for f2 in operon.get(feature2_name, regex):
            if f1 is f2:
                continue
            distance = _feature_distance(f1, f2)
            distances.append(distance)
    if not distances:
        return False
    if closest_pair_only:
        return min(distances) <= distance_bp
    return all([distance <= distance_bp for distance in distances])


def _calculate_all_distances(operon: Operon, feature: Feature) -> Optional[int]:
    """ Calculates the distance of a unique feature to all other features. """
    distances = []
    for other_feature in operon:
        if feature is other_feature:
            continue
        distance = _feature_distance(feature, other_feature)
        distances.append(distance)
    return distances


def _at_least_n_bp_from_anything(operon: Operon, feature_name: str, distance_bp: int, regex: bool = False) -> bool:
    """ Whether a given feature is more than distance_bp base pairs from another Feature. """
    at_least_one_good = False
    for feature in operon.get(feature_name, regex):
        distances = _calculate_all_distances(operon, feature)
        if not distances:
            return True
        if min(distances) < distance_bp:
            return False
        at_least_one_good = True
    return at_least_one_good


def _at_most_n_bp_from_anything(operon: Operon, feature_name: str, distance_bp: int, regex: bool = False) -> bool:
    """ Whether a given feature is less than distance_bp base pairs from any other feature. """
    at_least_one_good = False
    for feature in operon.get(feature_name, regex):
        distances = _calculate_all_distances(operon, feature)
        if not distances:
            return False
        if min(distances) > distance_bp:
            return False
        at_least_one_good = True
    return at_least_one_good


def _same_orientation(operon: Operon, exceptions: Optional[List[str]]) -> bool:
    """ Whether every gene is transcribed in the same direction. """
    exceptions = [] if exceptions is None else exceptions
    strands = set([feature.strand for feature in operon if feature.name.lower() not in map(str.lower, exceptions)])
    return len(strands) == 1


def _contains_features(operon: Operon, feature_names: List[str]) -> bool:
    """ Whether an Operon contains features with the given names. """
    return len(set(map(str.lower, operon.feature_names)) & set(map(str.lower, feature_names))) == len(feature_names)


def _contains_any_set_of_features(operon: Operon, sets: List[List[str]]) -> bool:
    """ Whether any the Operon has features with any of the given sets of names. """
    return any([_contains_features(operon, feature_names) for feature_names in sets])


def _contains_exactly_one_of(operon: Operon, f1: str, f2: str) -> bool:
    """ Whether the operon has one feature or another, but not both. """
    reg1 = re.compile(f1, re.IGNORECASE)
    reg2 = re.compile(f2, re.IGNORECASE)
    return any(map(reg1.match, operon.feature_names)) ^ any(map(reg2.match, operon.feature_names))


def _contains_at_least_n_features(operon: Operon, feature_names: List[str], feature_count: int, count_multiple_copies: bool) -> bool:
    """ Whether the operon has at least feature_count given features. """
    matches = [feature_name for feature_name in operon.feature_names if feature_name.lower() in map(str.lower, feature_names)]
    if len(matches) >= feature_count and count_multiple_copies:
        return True
    elif len(set(matches)) >= feature_count:
        return True
    else:
        return False


def _feature_distance(f1: Feature, f2: Feature) -> int:
    """ Returns the distance between two Features in base pairs. """
    distance1 = f2.start - f1.end
    distance2 = f1.start - f2.end
    # In the case of overlapping features, the distance is defined as 0
    return max(distance1, distance2, 0)


def _contains_group(operon: Operon, feature_names: List[str], max_gap_distance_bp: int, require_same_orientation: bool) -> bool:
    """ Determines whether Features with exactly the names in feature_names occur in any order without interruption by other Features, with gaps between each Feature no larger than max_gap_distance_bp. Optionally, every Feature can be required to be in the same orientation. """
    assert len(feature_names) > 1
    if len(operon) < len(feature_names):
        return False

    sorted_feature_names = tuple(sorted(map(str.lower, feature_names)))
    for operon_chunk in more_itertools.windowed(operon, len(feature_names)):
        sorted_chunk_names = tuple(sorted([feature.name.lower() for feature in operon_chunk]))
        if sorted_chunk_names == sorted_feature_names:
            max_gap_distance_in_group = 0
            for feature1, feature2 in zip(operon_chunk, operon_chunk[1:]):
                max_gap_distance_in_group = max(_feature_distance(feature1, feature2), max_gap_distance_in_group)
            if max_gap_distance_in_group > max_gap_distance_bp:
                continue
            if require_same_orientation:
                strands = set((feature.strand for feature in operon_chunk if feature.strand is not None))
                if len(strands) != 1:
                    continue
            return True
    return False


def _minimum_size(operon: Operon, feature_name: str, min_bp: int, all_matching_features_must_pass: bool, regex: bool) -> bool:
    results = [len(feature) >= min_bp for feature in operon.get(feature_name, regex)]
    if all_matching_features_must_pass:
        return all(results)
    return any(results)


def _maximum_size(operon: Operon, feature_name: str, max_bp: int, all_matching_features_must_pass: bool, regex: bool) -> bool:
    results = [len(feature) <= max_bp for feature in operon.get(feature_name, regex)]
    if all_matching_features_must_pass:
        return all(results)
    return any(results)



# Identifies inverted repeats and converts them to Feature objects
import re
import subprocess
import tempfile
from collections import namedtuple
from typing import Iterator, Optional, Tuple

from Bio.Seq import Seq

from operon_analyzer import genes, load

Repeat = namedtuple('Repeat', ['upstream_sequence', 'upstream_start', 'downstream_sequence', 'downstream_start'])
GRFResult = namedtuple('GRFResult', ['start', 'end', 'alignment'])
alignment_regex = re.compile(r'(?P<count>\d+)(?P<kind>\w)')


class BufferedSequence(object):
    """ Provides access to an operon's sequence, with a buffer flanking each side. This allows
    us to search for inverted repeats adjacent to a putative operon and get their exact sequences.
    All coordinates are internally converted to Python indicies so that we can easily slice sequences correctly,
    but we report the 1-based indexes to stay compatible with the rest of the library.
    """

    def __init__(self, operon: genes.Operon, contig_sequence: Seq, buffer_size: int):
        self._operon_start, self._operon_end = operon.feature_region
        self._contig_sequence = contig_sequence
        self._buffer_size = buffer_size

    @property
    def start(self) -> int:
        """ Returns the 0-based index of the beginning of the buffered sequence. """
        return max(0, self._operon_start - self._buffer_size)

    @property
    def end(self) -> int:
        """ Returns the 0-based, exclusive index of the end of the buffered sequence. """
        return min(len(self._contig_sequence), self._operon_end + self._buffer_size)

    @property
    def sequence(self) -> str:
        return self._contig_sequence[self.start:self.end]

    def __getitem__(self, key) -> str:
        """ Get slices of the contig sequence using coordinates that define the 0 position as the
        beginning of the buffered sequence, NOT the contig sequence. Used to get the sequence of
        the inverted repeats. """
        return self.sequence[key.start: key.stop]

    @property
    def operon_length(self) -> int:
        """ Gives the length of the putative operon. """
        return self._operon_end - self._operon_start + 1


def find_inverted_repeats(operon: genes.Operon, buffer_around_operon: int, min_repeat_size: int):
    """
    Searches an operon and the DNA flanking it for inverted repeats using :program:`GenericRepeatFinder`. 
    If found, they will be added to the operon as Feature objects with the name "TIR" and the sequence. 
    The strand will be set to 1 for the upstream sequence and -1 for the downstream sequence.

    Args:
        operon (Operon): The :class:`operon_analyzer.genes.Operon` object.
        buffer_around_operon (int): The number of base pairs on either side of the operon to search in addition to the operon's 
            internal sequence.
        min_repeat_size (int): The minimum number of base pairs that an inverted repeat must have.
    """
    contig_sequence = load.load_sequence(operon)
    _find_inverted_repeats(operon, contig_sequence, buffer_around_operon, min_repeat_size)


def _find_inverted_repeats(operon: genes.Operon, contig_sequence: Seq, buffer_size: int, min_repeat_size: int):
    """ Performs the inverted repeat search on a given DNA sequence. """
    buffered_sequence = BufferedSequence(operon, contig_sequence, buffer_size)

    # We are looking for systems where the inverted repeats are outside of the operon we're interested in, so we set a
    # spacer length that precludes finding both inside of the bounds of the operon's features. It's still possible for
    # one to be inside and one to be outside though.
    perfects, imperfects = _run_grf(buffered_sequence, min_repeat_size)

    # Go through each result, convert it to a pair of Features, and add the Features to the Operon
    n = 0  # must set in case we don't have any perfect results
    for n, result in enumerate(_parse_grf_results(perfects)):
        upstream_feature, downstream_feature = _parse_repeats(result, buffered_sequence, n)
        operon._features.append(upstream_feature)
        operon._features.append(downstream_feature)

    count = n
    for n, result in enumerate(_parse_grf_results(imperfects)):
        upstream_feature, downstream_feature = _parse_repeats(result, buffered_sequence, n + count + 1)
        operon._features.append(upstream_feature)
        operon._features.append(downstream_feature)


def _parse_repeats(result: GRFResult, buffered_sequence: BufferedSequence, number: int):
    # Extract the raw sequences of the inverted repeats and convert them to an alignment string (i.e. add gaps for deletions/insertions)
    upstream_size, downstream_size = _parse_alignment_size(result.alignment)
    raw_upstream_seq = str(buffered_sequence[result.start: result.start + upstream_size])
    raw_downstream_seq = str(buffered_sequence[result.end - downstream_size: result.end].reverse_complement())
    upstream_alignment, downstream_alignment = _format_aligned_sequences(raw_upstream_seq, raw_downstream_seq, result.alignment)

    # Figure out where the repeats are using the original coordinate system and create Feautre objects for each repeat
    upstream_feature = _make_tir_feature(buffered_sequence.start + result.start, upstream_size, raw_upstream_seq, upstream_alignment, 1, number)
    downstream_feature = _make_tir_feature(buffered_sequence.start + result.end - downstream_size, downstream_size, raw_downstream_seq, downstream_alignment, -1, number)
    return upstream_feature, downstream_feature


def _make_tir_feature(start: int, size: int, seq: str, alignment: str, strand: int, number) -> genes.Feature:
    """ Converts an inverted repeat into a Feature object. """
    assert len(seq) > 0
    assert start >= 0
    assert size > 0
    assert strand in (-1, 1)
    return genes.Feature(f"IR #{number} ({len(alignment)} bp)",
                         (start, start + size),
                         '',
                         strand,
                         '',
                         None,
                         f'{seq} {alignment}',
                         seq,
                         None)


def _run_grf(buffered_sequence: BufferedSequence, min_repeat_size: int) -> Optional[Tuple[str, str]]:
    """
    Runs GenericRepeatFinder and returns the raw text of the perfect and imperfect spacers that are found.

    sequence: the sequence of the putative operon with buffers on each side
    min_repeat_size: the smallest inverted or direct repeat allowable. Must be >= 5
    """
    assert min_repeat_size >= 5, "min_repeat_size must be at least 5"

    with tempfile.NamedTemporaryFile('w') as contig_file, tempfile.TemporaryDirectory() as grf_dir:
        contig_file.write(f">sequence\n{buffered_sequence.sequence}")
        # move the cursor to the beginning of the file or else we won't read any data from the temp file
        contig_file.seek(0)
        command = ["grf-main",
                   "-i", contig_file.name,
                   "-o", grf_dir,
                   "-c", "0",
                   "-f", "1",
                   "-s", str(min_repeat_size),
                   "--min_tr", str(min_repeat_size),
                   "--min_space", str(buffered_sequence.operon_length),
                   "--max_space", str(len(buffered_sequence.sequence))]
        result = subprocess.run(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if result.returncode != 0:
            return None
        with open(f"{grf_dir}/perfect.spacer.id") as f:
            perfect = f.read().strip()
        with open(f"{grf_dir}/imperfect.id") as f:
            imperfect = f.read().strip()
        return perfect, imperfect


def _parse_grf_results(raw_text: str) -> Iterator[GRFResult]:
    """ Parses the raw text of GenericRepeatFinder output. Assumes that only
    IDs are written to the file, sequences should be omitted. If no results
    were found, the text will be empty. """
    if not raw_text:
        return None
    for line in raw_text.split("\n"):
        yield _parse_repeat_id(line)


def _parse_repeat_id(repeat_id: str) -> Tuple[int, int, str]:
    """
    Parses one line from GenericRepeatFinder output.
    start is the location of the beginning of the upstream repeat.
    end is the location of the end of the downstream repeat.
    alignment is a serialization of the difference between the two sequences
    (see _parse_alignment_size for details).
    """
    _, start, end, alignment = repeat_id.split(":")
    # GRF uses 1-based indexes, with inclusive ends. Subtracting 1 from the start
    # coordinate allows us to use the coordinates as slice indices
    result = GRFResult(int(start) - 1, int(end), alignment)
    assert result.start >= 0
    return result


def _parse_alignment_size(alignment: str) -> Tuple[int, int]:
    """
    We are given the difference (if any) between the upstream and downstream
    repeat in a serialized code. In order to determine the actual sequence of the repeat,
    we have to parse this string just to determine the size of the repeats.

    The letters mean:
    m:  perfect match
    M:  aligned mismatch
    I:  insertion in upstream repeat
    D:  deletion in upstream repeat

    For example, `5m3I16m` would be a repeat with 5 homologous base pairs, 3 insertions
    in the upstream repeat relative to the downstream repeat, followed by 16 homologous
    base pairs.
    """
    sizes = {'m': (1, 1), 'M': (1, 1), 'I': (1, 0), 'D': (0, 1)}
    matches = alignment_regex.findall(alignment)
    upstream_total, downstream_total = 0, 0
    for count, kind in matches:
        count = int(count)
        upfactor, downfactor = sizes[kind]
        upstream_total += upfactor * count
        downstream_total += downfactor * count
    return upstream_total, downstream_total


def _format_aligned_sequences(raw_upstream_sequence: str,
                              raw_downstream_sequence: str,
                              alignment: str) -> Tuple[str, str]:
    """ Takes the raw sequences of each repeat and inserts dashes to indicate gaps, as appropriate. """
    matches = alignment_regex.findall(alignment)
    upseq = []
    downseq = []
    up_position = 0
    down_position = 0
    for count, kind in matches:
        count = int(count)
        if kind in ('m', 'M'):
            # mismatches don't require a dash
            upseq.append(raw_upstream_sequence[up_position: up_position + count])
            downseq.append(raw_downstream_sequence[down_position: down_position + count])
            up_position += count
            down_position += count
        elif kind == 'I':
            # insertions mean the upstream repeat has an extra base relative to the downstream repeat
            upseq.append(raw_upstream_sequence[up_position: up_position + count])
            downseq.append('-' * count)
            up_position += count
        elif kind == 'D':
            # deletions mean the downstream repeat has an extra base relative to the upstream repeat
            upseq.append('-' * count)
            downseq.append(raw_downstream_sequence[down_position: down_position + count])
            down_position += count
        else:
            raise ValueError(f"unexpected alignment value: {alignment}")
    # convert the lists that are accumulating strings to a single aligned sequence
    return "".join(upseq), "".join(downseq)



from gene_finder.orffinder import orffinder, neighborhood_orffinder
from gene_finder.utils import concatenate
from gene_finder.steps import (SearchStep, 
                                FilterStep, 
                                SeedStep,
                                SeedWithCoordinatesStep, 
                                CrisprStep, 
                                Blastn,
                                BlastnStep,
                                Blastp, 
                                Blastpsi, 
                                MMseqs, 
                                Diamond,
                                Pilercr)

from gene_finder.output_writers import CSVWriter

import tempfile, os, json, gzip
from Bio import SeqIO

class Pipeline:
    """
    Coordinates protein (or nucleic acid) searches to find gene clusters
    of interest in genomic/metagenomic data.
    """

    def __init__(self):
        """
        Create a pipeline for finding gene cluster candidates.
        """
        # Most members are re-assigned later, either when search steps are added 
        # or when Pipeline.run() is called, but each is intialized here for clarity

        # defined in Pipeline.run()
        self.data_path = None
        self.min_prot_len = None
        self.span = None
        self.job_id = None
        self.output_directory = None
        
        # modified by add_step methods
        self._steps = []
        
        # collect results about the input data
        # will be reset each time Pipeline.run() is called
        self._results = {}
        self._all_hits = {}

        # basically give the current state of the output files; that is, 
        # whether any data from this run has been written to disk yet
        # these are also reset (to this state) when Pipeline.run() is called
        self._appending_results = False
        self._appending_hits = False

        # collect data specific to a single contig in the input
        # will be reset each time a new contig is processed
        self._working_results = {}
        self._all_orfs = None
        self._neighborhood_orfs = {}
        self._working_dir = None

    
    def __del__(self):
        """
        Delete the working directory and its contents when this object is garbage collected.
        """
        if self._working_dir is not None:
            self._working_dir.cleanup()
    

    def _setup_run(self, record):
        """
        Prepare for running the pipeline on a new contig. Since the same setup is re-used on 
        multple contigs (or even multiple fasta files) we want to make sure that old data is 
        cleared out.

        Also gather info about the current contig and call `orffinder` to get all ORFs in
        this contig.
        """
        self._reset_contig_data()
        self._working_dir = tempfile.TemporaryDirectory()
        contig_id = record.id
        contig_len = len(record.seq)
        contig_path = os.path.join(self._working_dir.name, "contig.fasta")
        SeqIO.write(record, contig_path, "fasta")
        self._get_all_orfs(contig_path, contig_id)
        self._all_hits[contig_id] = {}
        return contig_id, contig_len, contig_path
    

    def _reset_contig_data(self):
        """
        Clear out data from a previous contig.
        """
        if self._working_dir is not None:
            self._working_dir.cleanup()
        self._working_results = {}
        self._neighborhood_orfs = {}
        self._all_orfs = None
    

    def _reset_results(self):
        """
        Clear out results from a previous run.
        """
        self._results = {}
        self._all_hits = {}
        self._appending_hits = False
        self._appending_results = False


    def _open_data(self, data, is_binary):
        """
        Return the appropriate handle/file object for reading
        input data that is either flat text or a gzipped file. 
        """
        if is_binary:
            handle = gzip.open(data, "rt")
        else:
            handle = open(data, "r")
        
        return handle


    def _results_init(self, neighborhood_ranges):
        """
        Create an entry for every gene neighborhood identified
        during the seed phase in the working results tracker.
        """
        for r in neighborhood_ranges:
            key = "Loc_{}-{}".format(r[0], r[1])
            self._working_results[key] = {"Loc_start-pos": int(r[0]), "Loc_end-pos": int(r[1]), 
                                    "new_hit_count": 0, "Hits": {}}

    
    def _filter(self, min_prot_count):
        """
        Remove neighborhoods from the working results if the number of
        new hits added during the filter step is less than the user 
        specified cutoff. If no cutoff is given the default used is 1.
        """
        remove = [neighborhood for neighborhood in self._working_results 
                    if self._working_results[neighborhood]["new_hit_count"] < min_prot_count]
        
        for neighborhood in remove:
            del self._working_results[neighborhood]
            del self._neighborhood_orfs[neighborhood]
    
    
    def _results_update(self, hits, min_prot_count):
        """
        Add hits to the results tracker (grouped by neighborhood).
        """
        for hit in hits:
            for neighborhood in self._working_results:

                # note that the begining and end of the hit is denoted by 
                # the begining and end of the orf used as the query
                h_start = min(int(hits[hit]["Query_start-pos"]), 
                                int(hits[hit]["Query_end-pos"]))
                h_stop = max(int(hits[hit]["Query_start-pos"]), 
                                int(hits[hit]["Query_end-pos"]))
                
                # check whether this hit is contained within the neighborhood
                if (h_start >= self._working_results[neighborhood]["Loc_start-pos"] 
                    and h_stop <= self._working_results[neighborhood]["Loc_end-pos"]):
                    self._working_results[neighborhood]["Hits"][hit] = hits[hit]
                    self._working_results[neighborhood]["new_hit_count"] += 1
        
        if min_prot_count >= 1:
            self._filter(min_prot_count)
        
        for neighborhood in self._working_results:
            self._working_results[neighborhood]["new_hit_count"] = 0

    
    def _results_update_crispr(self, hits):
        """
        Add hits for CRISPR arrays to the results tracker.
        """
        for hit in hits:
            for neighborhood in self._working_results:

                h_start = int(hits[hit]["Position"])
                if (h_start >= self._working_results[neighborhood]["Loc_start-pos"] 
                    and h_start <= self._working_results[neighborhood]["Loc_end-pos"]):
                    self._working_results[neighborhood]["Hits"][hit] = hits[hit]

    
    def _results_update_nucl(self, hits):
        """
        Add hits for nucleotide BLAST.
        """
        for hit in hits:
            for neighborhood in self._working_results:

                # note that the begining and end of the hit is denoted by 
                # the begining and end of the orf used as the query
                h_start = min(int(hits[hit]["Query_start-pos"]), 
                                int(hits[hit]["Query_end-pos"]))
                h_stop = max(int(hits[hit]["Query_start-pos"]), 
                                int(hits[hit]["Query_end-pos"]))
                
                # check whether this hit is contained within the neighborhood
                if (h_start >= self._working_results[neighborhood]["Loc_start-pos"] 
                    and h_stop <= self._working_results[neighborhood]["Loc_end-pos"]):
                    self._working_results[neighborhood]["Hits"][hit] = hits[hit]
                    self._working_results[neighborhood]["new_hit_count"] += 1
        
        for neighborhood in self._working_results:
            self._working_results[neighborhood]["new_hit_count"] = 0


    def _get_all_orfs(self, data, id):
        """
        Get all of the (translated) open reading frames in this genome.
        """
        orfs = os.path.join(self._working_dir.name, "all_orfs.fasta")
        self._all_orfs = orffinder(sequence=data, output=orfs, 
                                    min_prot_len=self.min_prot_len, description=id)

    
    def _get_orfs_in_neighborhood(self, ranges, data, id):
        """
        Grab all of the open reading frames within a subsequence
        (neighborhood) from the original parent.   
        """
        self._neighborhood_orfs = {}
        neighborhood_orffinder(sequence=data, ranges=ranges, 
                                outdir=self._working_dir.name, 
                                min_prot_len=self.min_prot_len, 
                                description=id)

        for r in ranges:
            key = "Loc_{}-{}".format(r[0], r[1])
            path = os.path.join(self._working_dir.name, 
                                "orf_{}_{}.fasta".format(r[0], r[1]))
            self._neighborhood_orfs[key] = path

    
    def add_seed_step(self, db, name, e_val, blast_type, sensitivity=None, parse_descriptions=True, blast_path=None, **kwargs):
        """
        Find genomic regions that contain at least one "seed" sequence.

        Args:
            db (str): Path to the target (seed) protein database.
            name (str): A unique name/ID for this step in the pipeline.
            e_val (float): Expect value to use. Only keep hits with a an equivalent or
                better (lower) score.
            blast_type (str): Specifies which search program to use. 
                This can be either "PROT" (blastp), "PSI" (psiblast),
                "mmseqs" (mmseqs2), or "diamond" (diamond).
            sensitivity (str): Sets the sensitivity param 
                for mmseqs and diamond (does nothing if BLAST is the
                seach type).
            parse_descriptions (bool, optional): By default, reference protein
                descriptions (from fasta headers) are parsed for gene name labels;
                specifically, descriptions are split on whitespace characters
                and the second item is used for the label. Make this false to 
                simply use the whole protein description for the label 
                (i.e everything after the first whitespace in the header). If
                using this option with NCBI BLAST, DO NOT use the ``-parse_seqids``
                flag when creating protein databases with :program:`makeblastdb`.
            blast_path (string, optional): Path to the blastp/mmseqs/diamond program,
                if not using the system default.
            **kwargs: These can be any additional BLAST parameters,
                specified as key-value pairs. Note that certain parameters
                are not allowed, mainly those that control output formatting.
                Currently only supported for blastp/psiblast; if blast_type
                is set to mmseqs or diamond, kwargs will be silently ignored.

        Note:
            This should be the first step added to a :class:`gene_finder.pipeline.Pipeline` object.
            Additional gene finding steps can be added in any order. 
        """
        if blast_type in ("PROT", "blastp"):
            path = 'blastp' if blast_path is None else blast_path
            self._steps.append(SeedStep(Blastp(db, e_val, name, parse_descriptions, path, kwargs)))
        elif blast_type in ("PSI", "psiblast"):
            path = 'psiblast' if blast_path is None else blast_path
            self._steps.append(SeedStep(Blastpsi(db, e_val, name, parse_descriptions, path, kwargs)))
        elif blast_type == "mmseqs":
            self._steps.append(SeedStep(MMseqs(db, str(e_val), name, sensitivity, parse_descriptions)))
        elif blast_type == "diamond":
            self._steps.append(SeedStep(Diamond(db, str(e_val), name, sensitivity, parse_descriptions)))
        else:
            raise ValueError("blast type option '{}' not recognized".format(blast_type))
    

    def add_seed_with_coordinates_step(self, db, name, e_val, blast_type, sensitivity=None, 
                                       parse_descriptions=True, start=None, end=None, 
                                       contig_id=None, blast_path=None, **kwargs):
        """
        Define a genomic region of interest with coordinates instead of a seed sequence.

        An alternative to :meth:`gene_finder.pipeline.Pipeline.add_seed_step`. Most useful for re-annotating 
        putative systems of interest, where the region coordinates are already
        known.

        Args:
            db (str): Path to the target database to search against.
            name (str): A unique name/ID for this step in the pipeline.
            e_val (float): Expect value to use. Only keep hits with an equivalent or
                better (lower) score.
            blast_type (str): Specifies which search program to use. 
                This can be either "PROT" (blastp), "PSI" (psiblast),
                "mmseqs" (mmseqs2), or "diamond" (diamond).
            sensitivity (str): Sets the sensitivity param 
                for mmseqs and diamond (does nothing if BLAST is the
                seach type).
            parse_descriptions (bool, optional): By default, reference protein
                descriptions (from fasta headers) are parsed for gene name labels;
                specifically, descriptions are split on whitespace characters
                and the second item is used for the label. Make this false to 
                simply use the whole protein description for the label 
                (i.e everything after the first whitespace in the header). If
                using this option with NCBI BLAST, DO NOT use the ``-parse_seqids``
                flag when creating protein databases with :program:`makeblastdb`.
            start (int): Defines the beginning of the region to search, in base pairs (bp).
                If no start position is given the first (zero indexed) position
                in the genome/contig is used.
            end (int): Defines the end of the region to search, in base pairs (bp). If no
                end position is given the last position in the contig is used.
            contig_id(string, optional): An identifier for the contig to search.
                If no ID is given, the pipeline will search every contig in the
                input file using the coordinates specified. Note that the contig ID 
                is defined as the substring between the ">" character and the first
                " " character in the contig header.
            blast_path (string, optional): Path to the blastp/mmseqs/diamond program,
                if not using the system default.
            **kwargs: These can be any additional BLAST parameters,
                specified as key-value pairs. Note that certain parameters
                are not allowed, mainly those that control output formatting.
                Currently only supported for blastp/psiblast; if blast_type
                is set to mmseqs or diamond, kwargs will be silently ignored.
        """
        self._steps.append(SeedWithCoordinatesStep(start=start, end=end, contig_id=contig_id))
        self.add_blast_step(db=db, name=name, e_val=e_val, blast_type=blast_type, 
                            sensitivity=sensitivity, parse_descriptions=parse_descriptions, blast_path=blast_path, **kwargs)


    def add_filter_step(self, db, name, e_val, blast_type, min_prot_count=1, 
                        sensitivity=None, parse_descriptions=True, blast_path=None, **kwargs):
        """
        Add a step to search candidate regions for target sequences, and filter out
        candidates that do not have at least ``min_prot_count`` matching sequences.

        Args:
            db (str): Path to the target protein sequence database.
            name (str): A unique name/ID for this step in the pipeline.
            e_val (float): Expect value to use. Only keep hits with a an equivalent or
                better (lower) score.
            blast_type (str): Specifies which search program to use. 
                This can be either "PROT" (blastp), "PSI" (psiblast),
                "mmseqs" (mmseqs2), or "diamond" (diamond).
            min_prot_count (int, optional): Minimum number of hits 
                needed to keep each candidate.
            sensitivity (str): Sets the sensitivity param 
                for mmseqs and diamond (does nothing if BLAST is the
                seach type).
            parse_descriptions (bool, optional): By default, reference protein
                descriptions (from fasta headers) are parsed for gene name labels;
                specifically, descriptions are split on whitespace characters
                and the second item is used for the label. Make this false to 
                simply use the whole protein description for the label 
                (i.e everything after the first whitespace in the header). If
                using this option with NCBI blast, DO NOT use the ``-parse_seqids``
                flag when creating protein databases with :program:`makeblastdb`.  
            blast_path (string, optional): Path to the blastp/mmseqs/diamond program,
                if not using the system default.
            **kwargs: These can be any additional BLAST parameters,
                specified as key-value pairs. Note that certain parameters
                are not allowed, mainly those that control output formatting.
                Currently only supported for blastp/psiblast; if blast_type
                is set to mmseqs or diamond, kwargs will be silently ignored.
        """
        if blast_type in ("PROT", "blastp"):
            path = 'blastp' if blast_path is None else blast_path
            self._steps.append(FilterStep(Blastp(db, e_val, name, parse_descriptions, path, kwargs), min_prot_count))
        elif blast_type in ("PSI", "psiblast"):
            path = 'psiblast' if blast_path is None else blast_path
            self._steps.append(FilterStep(Blastpsi(db, e_val, name, parse_descriptions, path, kwargs), min_prot_count))
        elif blast_type == "mmseqs":
            self._steps.append(FilterStep(MMseqs(db, str(e_val), name, sensitivity, parse_descriptions), min_prot_count))
        elif blast_type == "diamond":
            self._steps.append(FilterStep(Diamond(db, str(e_val), name, sensitivity, parse_descriptions), min_prot_count))
        else:
            raise ValueError("blast type option '{}' not recognized".format(blast_type))
    
    
    def add_blast_step(self, db, name, e_val, blast_type, 
                        sensitivity=None, parse_descriptions=True, blast_path=None, **kwargs):
        """
        Add a non-filtering search step to the pipeline. That is, search each candidate 
        for target sequences without applying any filtering logic. This is most useful
        for annotating candidates for non-essential or ancillary genes.

        Args:
            db (str): Path to the target protein sequence database.
            name (str): A unique name/ID for this step in the pipeline.
            e_val (float): Expect value to use. Only keep hits with a an equivalent or
                better (lower) score.
            blast_type (str): Specifies which search program to use. 
                This can be either "PROT" (blastp), "PSI" (psiblast),
                "mmseqs" (mmseqs2), or "diamond" (diamond).
            sensitivity (str): Sets the sensitivity param 
                for mmseqs and diamond (does nothing if BLAST is the
                seach type). 
            parse_descriptions (bool, optional): By default, reference protein
                descriptions (from fasta headers) are parsed for gene name labels;
                specifically, descriptions are split on whitespace characters
                and the second item is used for the label. Make this false to 
                simply use the whole protein description for the label 
                (i.e everything after the first whitespace in the header). If
                using this option with NCBI BLAST, DO NOT use the ``-parse_seqids``
                flag when creating protein databases with :program:`makeblastdb`.
            blast_path (string, optional): Path to the blastp/mmseqs/diamond program,
                if not using the system default.
            **kwargs: These can be any additional BLAST parameters,
                specified as key-value pairs. Note that certain parameters
                are not allowed, mainly those that control output formatting.
                Currently only supported for blastp/psiblast; if blast_type
                is set to mmseqs or diamond, kwargs will be silently ignored.     
        """
        if blast_type in ("PROT", "blastp"):
            path = 'blastp' if blast_path is None else blast_path
            self._steps.append(SearchStep(Blastp(db, e_val, name, parse_descriptions, path, kwargs)))
        elif blast_type in ("PSI", "psiblast"):
            path = 'psiblast' if blast_path is None else blast_path
            self._steps.append(SearchStep(Blastpsi(db, e_val, name, parse_descriptions, path, kwargs)))
        elif blast_type == "mmseqs":
            self._steps.append(SearchStep(MMseqs(db, str(e_val), name, sensitivity, parse_descriptions)))
        elif blast_type == "diamond":
            self._steps.append(SearchStep(Diamond(db, str(e_val), name, sensitivity, parse_descriptions)))
        else:
            raise ValueError("blast type option '{}' not available for filter step".format(blast_type))
    
    
    def add_crispr_step(self):
        """
        Add a step to search for CRISPR arrays using PILER-CR.
        """
        self._steps.append(CrisprStep(Pilercr("CRISPR")))
    

    def add_blastn_step(self, db, name, e_val, parse_descriptions=False, blastn_path='blastn', **kwargs):
        """ 
        Add a step to do nucleotide BLAST. 

        Args:
            db (str): Path to the target protein sequence database.
            name (str): A unique name/ID for this step in the pipeline.
            e_val (float): Expect value to use. Only keep hits with a an equivalent or
                better (lower) score.
            parse_descriptions (bool, optional): By default, reference protein
                descriptions (from fasta headers) are parsed for gene name labels;
                specifically, descriptions are split on whitespace characters
                and the second item is used for the label. Make this false to 
                simply use the whole protein description for the label 
                (i.e everything after the first whitespace in the header). If
                using this option with NCBI BLAST, DO NOT use the ``-parse_seqids``
                flag when creating protein databases with :program:`makeblastdb`.
            blast_path (string, optional): Path to the blastn program,
                if not using the system default.
            **kwargs: These can be any additional BLAST parameters,
                specified as key-value pairs. Note that certain parameters
                are not allowed, mainly those that control output formatting.
                Currently only supported for blastp/psiblast; if blast_type
                is set to mmseqs or diamond, kwargs will be silently ignored.
        """
        self._steps.append(BlastnStep(Blastn(db, name, e_val, parse_descriptions, blastn_path, kwargs)))


    def _update_output_sequences(self):
        """
        Replace alignment sequences (which contain gap characters) with full ORF sequences 
        in the final output.
        """
        for neighborhood, path in self._neighborhood_orfs.items():
            sequences = {}
            for record in SeqIO.parse(path, "fasta"):
                sequences[record.id] = str(record.seq)
            for key, hit in self._working_results[neighborhood]["Hits"].items():
                if "Query_ORFID" in hit.keys():
                    self._working_results[neighborhood]["Hits"][key]["Query_seq"] = sequences[hit["Query_ORFID"]]

    
    def _format_results(self, results_data, incremental_output):
        """
        Process results into their final CSV format and write them to disk.
        """
        # Remove temporary hit counter tag
        for contig in results_data:
            for neighborhood in results_data[contig]:
                del results_data[contig][neighborhood]["new_hit_count"]
        
        filename = "{}_results.csv".format(self.job_id) if self.job_id is not None else "gene_finder_results.csv"
        if self.output_directory is not None and os.path.exists(self.output_directory):
            filename = os.path.join(self.output_directory, filename)
        csv_writer = CSVWriter(results_data, filename)
        mode = "a" if incremental_output and self._appending_results else "w"
        csv_writer.to_csv(self.data_path, mode)
        self._appending_results = True
    
    
    def _record_all_hits(self, all_hit_data):
        """
        Write all hits from one or more contigs to disk (in json format).
        """
        filename = "{}_hits.json".format(self.job_id) if self.job_id is not None else "gene_finder_hits.json"
        if self.output_directory is not None and os.path.exists(self.output_directory):
            filename = os.path.join(self.output_directory, filename)
        mode = "a" if self._appending_hits else "w"
        # looping allows the json-like data to be written incrementally, which is 
        # admittedly not really what json is intended for
        # we'll do this regardless of whether `incremental_output` is true so
        # that the file format is the same in both cases
        with open(filename, mode) as f:
            for contig in all_hit_data:
                final_candidate_count = self._final_candidate_count(contig)
                contig_data = {contig: {"final_candidate_count": final_candidate_count, "hits": all_hit_data[contig]}}
                json.dump(contig_data, f)
                f.write("\n")
        self._appending_hits = True


    def _final_candidate_count(self, contig_id):
        """
        Returns the number of candidate systems (per contig) that were found.
        """
        return len(self._results[contig_id])
    
    
    def _write_checkpoint_file(self, contig_id):
        """
        Write the ID of the current contig (i.e the contig currently being processed)
        to disk. This gets overwritten with each new contig ID, and removed once the
        job completes successfully. The point is that if the user is running gene 
        finder in "incremental" mode and the job fails unexpectedly, the ID can be
        used to re-start gene finder from where it left off.
        """
        filename = "{}_checkpoint.txt".format(self.job_id) if self.job_id is not None else "gene_finder_checkpoint.txt"
        if self.output_directory is not None and os.path.exists(self.output_directory):
            filename = os.path.join(self.output_directory, filename)
        with open(filename, "w") as f:
            f.write("{},{}".format(contig_id, self.data_path))
    

    def _remove_checkpoint_file(self, incremental_output):
        """
        Remove the contig ID checkpoint file. This is only called after the entire job
        has been successfully completed.
        """
        filename = "{}_checkpoint.txt".format(self.job_id) if self.job_id is not None else "gene_finder_checkpoint.txt"
        if self.output_directory is not None and os.path.exists(self.output_directory):
            filename = os.path.join(self.output_directory, filename)
        # check that the checkpoint file was actually created during this run
        if os.path.exists(filename) and incremental_output:
            os.remove(filename)
    

    def run(self, data, job_id=None, output_directory=None, min_prot_len=60, 
            span=10000, record_all_hits=False, incremental_output=False,
            starting_contig=None, gzip=False) -> dict:
        """
        Execute each step in the pipeline, in the order they were added.

        Args:
            data (str): Path to the input data file. Can be a single-
                or multi-sequence file in fasta format.
            job_id (str, optional): A unique ID to prefix all output
                files. If no ID is given, the string "gene_finder" 
                will be used as the prefix. In any case, results from
                the pipeline are written to the file <prefix>_results.csv.
            output_directory (str, optional): The directory to write
                output data files to. If no directory is given then the current
                (working) directory is used.
            min_prot_len (int, optional): Minimum ORF length (aa).
                Default is 60.
            span (int, optional): Length (nt) upsteam and downstream
                of each seed hit to keep. Defines the aproximate size
                of the genomic neighborhoods that will be used as the
                search space after the seed step.
            record_all_hits (bool, optional): Write data about all genes found 
                (even discarded ones) to the file <job_id>_hits.json,
                grouped by contig. Note that this contains much of the same
                information as is in the results CSV file; nevertheless, it
                may be useful for analysis or troubleshooting a search.
            incremental_output (bool, optional): Write results to disk
                after each contig is processed. Using this option also creates a 
                checkpoint file that gives the ID of the contig that is currently 
                being processed; if the job finishes successfully, this file will 
                be automatically cleaned up. This feature is especially useful 
                for long-running jobs. 
            starting_contig (bool, optional): The sequence identifier of
                the contig where the run should begin. In other words, 
                skip over records in the input file until
                the specified contig is reached, and then run the pipeline
                as normal. This is usually used in conjunction with 
                ``incremental_output``.
            gzip (bool, optional): Was this file compressed with gzip? 

        Returns:   
            dict: Candidate systems, grouped by contig id and genomic location.
        """
        self._reset_results()
        self.data_path = data
        self.min_prot_len = min_prot_len
        self.span = span
        self.job_id = job_id
        self.output_directory = output_directory

        data_handle = self._open_data(self.data_path, gzip)
        for record in SeqIO.parse(data_handle, "fasta"):
            if starting_contig is not None:
                if record.id == starting_contig:
                    starting_contig = None
                else:
                    continue
            # if this run was seeded with a particular contig id, skip all other contigs
            # in the file if they exit
            # saves a bit of time, since there is overhead associated with getting things set up
            if isinstance(self._steps[0], SeedWithCoordinatesStep):
                if self._steps[0].contig_id is not None and self._steps[0].contig_id != record.id:
                    continue
            # clear out any data that may be leftover from processing a previous
            # contig and get all ORFs in this contig
            contig_id, contig_len, contig_path = self._setup_run(record)
            if incremental_output:
                self._write_checkpoint_file(contig_id)
            if self._all_orfs is None:
                # No ORFs were identified in the contig (probably because it's small),
                # so we just continue on to the next contig if it exists
                self._all_hits[contig_id] = {}
                self._results[contig_id] = {}
                if incremental_output and record_all_hits:
                    self._record_all_hits({contig_id: {}})
                continue
            for step in self._steps:
                if isinstance(step, SeedStep):
                    #print("Begin seed step: {}".format(step.name))
                    step.execute(self._all_orfs, self.span, contig_len)
                    self._get_orfs_in_neighborhood(step.neighborhood_ranges, 
                                                    contig_path,
                                                    contig_id)
                    self._results_init(step.neighborhood_ranges)
                    self._results_update(step.hits, min_prot_count=0)
                    neighborhood_orfs = concatenate(self._working_dir.name, 
                                                    self._neighborhood_orfs.values())

                elif isinstance(step, SeedWithCoordinatesStep):
                    if step.start is None:
                        step.update_start_coord(0)
                    if step.end is None:
                        step.update_end_coord(contig_len)
                    self._get_orfs_in_neighborhood(step.neighborhood_ranges,
                                                   contig_path,
                                                   contig_id)
                    self._results_init(step.neighborhood_ranges)
                    neighborhood_orfs = concatenate(self._working_dir.name,
                                                    self._neighborhood_orfs.values())
                    # skip hit processing at the end of the loop, since we don't have hits yet
                    continue

                elif isinstance(step, FilterStep):
                    #print("Begin filter step: {}".format(step.name))
                    step.execute(neighborhood_orfs)
                    self._results_update(step.hits, min_prot_count=step.min_prot_count)
                    neighborhood_orfs = concatenate(self._working_dir.name, 
                                                    self._neighborhood_orfs.values())
                    #print("Filtering for {} genes complete".format(step.name))
                
                elif isinstance(step, CrisprStep):
                    #print("Begin CRISPR array search")
                    step.execute(contig_path)
                    self._results_update_crispr(step.hits)
                    #print("CRISPR array search complete")
                        
                elif isinstance(step, BlastnStep):
                    step.execute(contig_path)
                    self._results_update_nucl(step.hits)

                else:
                    #print("Begin blast step: {}".format(step.name))
                    step.execute(neighborhood_orfs)
                    self._results_update(step.hits, min_prot_count=0)
                    #print("Blast step complete")
                
                if len(step.hits) != 0:
                    self._all_hits[contig_id][step.search_tool.step_id] = step.hits
                # check if all candidates have been eliminated (if so, no need to
                # continue searching)
                if len(self._neighborhood_orfs) == 0:
                    #print("No putative neighborhoods remain - terminating run")
                    self._results[contig_id] = {}
                    break
                self._update_output_sequences()
                self._results[contig_id] = self._working_results

            # The current contig is finished processing; if outputing results incrementally,
            # we'll dump data about this contig to disk now 
            if incremental_output:
                self._format_results({contig_id: self._results[contig_id]}, True)
                if record_all_hits:
                    self._record_all_hits({contig_id: self._all_hits[contig_id]})

        # all contigs have finished processing; dump all data to disk if we haven't already
        if not incremental_output:
            self._format_results(self._results, False)
            if record_all_hits:
                self._record_all_hits(self._all_hits)
        
        self._remove_checkpoint_file(incremental_output)
        data_handle.close() # make sure to close the input data file object
        return self._results



import setuptools

setuptools.setup(
    name="opfi", 
    version="0.1.1",
    author="Alexis M Hill, James Rybarski",
    author_email="alexismhill3@gmail.com",
    description="A package for discovery, annotation, and analysis of gene clusters in genomics or metagenomics datasets.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/wilkelab/Opfi",
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "biopython==1.76", 
        "pytest==5.3.2",
        "hypothesis==5.1.1", 
        "matplotlib==3.2.1", 
        "PyYAML==5.4", 
        "dna-features-viewer==3.0.1", 
        "more-itertools==8.4.0", 
        "parasail==1.2", 
        ]
)
